/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/*

  This file generates the dynamically
  generated classes. In order to understand what is going on,
  you first need to read about the ASM framework, on which
  PolyD relies.

  The home page of PolyD is http://www.ovmj.org/polyd
  The home page of ASM   is http://asm.objectweb.org


  The code that needs to be constructed is rather complicated,
  and that reflects the number of modifiers that can affect
  the operation of the new dispatcher. For example, the @IfNull
  tag causes an additional test to be inserted. The @Raw tag
  leads to a different handling of the corresponding argument.
  The @As tag leads to the class constant to be used rather
  than the dynamic class of the argument. And so on.

  All these modifiers lead to a number of high number of
  localized "ifs" that alter the behaviour of tiny bits of
  code. The fact that several features were added over
  time also made the code probably a lot more complicated
  than what it should really be. For these reasons, the
  code below is really not nice, and would probably benefit
  from a massive refactoring.

  Nonetheless, even as it is, most of the code is relatively
  easy to understand after checking an example of the code
  that is generated. An example is at the very end of this
  file, it does not show everything, but the main logic is
  there.

  The dynamically generated class contains the following
  components:
  - static fields that store the hashmaps that form the
    method caches. There is a group formed by one hashmap
    and its arrays for each MethodGroup, that is for
    each prototype in the interface.
  - one method for each prototype in the interface. The
    method implements the actual dynamic selection.
    Most of the times it can just check into a couple of
    arrays and find out the correct method that should
    be delegated to execution, but sometimes it needs
    to use the MultiHash for more complex operations.
  - one special static initializer that prepares the
    hashmaps

  Each class is supplemented by a dynamically
  generated factory, which does:
  - nothing more than instantiating the dispatcher, storing
    the bodies corresponding to the implementing classes
    in some non-static fields and setting  up the "self"
    fields, if necessary.

*/

// for examples of the dynamically generated code,
// go to the end of this file

package ovm.polyd;

import java.io.FileOutputStream;
import java.lang.reflect.*;

import java.util.List;
import java.util.Vector;
import ovm.polyd.Missing;


import org.objectweb.asm.*;

import static ovm.polyd.MethodGroup.DispatchInfo.*;
import ovm.polyd.Utils.Diff;
import static ovm.polyd.Utils.Diff.*;

// Some complications are introduced by the
// presence of the @Raw tag. While precedently it could
// always be assumed that the prototype and the called
// implementations were the same length, it can now be that
// the prototype has MORE parameters. Worse, for some
// data structures it makes sense to keep a list that is
// aligned with the list of parameters of the implementations,
// while for others the list should be aligned with the list
// of parameters in the prototype. The distinction is important
// and should be kept into account at all times while touching
// the code.

class ClassBuilder extends ClassLoader implements Constants {
 // Java 1.5 is required for LDC of "class" constants
 // define as true to dynamically generate 1.5 bytecode

 // note that the macro comment ifdef14 really means ifNotJava15

/*#ifdef14
 private static final boolean java15=false;
#else14 */
 private static final boolean java15=true;
//#endif14

 private static String thisPackage="ovm.polyd".replace('.','/').concat("/");

 private static void genReturn(CodeVisitor mw,Class cc) {
  if (cc.isPrimitive()) {
   if (cc==void.class)
    mw.visitInsn(RETURN);
   else if(cc==byte.class)
    mw.visitInsn(IRETURN);
   else if(cc==char.class)
    mw.visitInsn(IRETURN);
   else if(cc==double.class)
    mw.visitInsn(DRETURN);
   else if(cc==float.class)
    mw.visitInsn(FRETURN);
   else if(cc==int.class)
    mw.visitInsn(IRETURN);
   else if(cc==long.class)
    mw.visitInsn(LRETURN);
   else if(cc==short.class)
    mw.visitInsn(IRETURN);
   else if(cc==boolean.class)
    mw.visitInsn(IRETURN);
   else {
    throw new InternalError(0x3C1,
     "Internal error: unknown class in genReturn()."  
    );
   }
  } else {
   mw.visitInsn(ARETURN);
  }
 }

 private static void genDummyReturn(CodeVisitor mw,Class cc) {
  if (cc.isPrimitive()) {
   if (cc==void.class) {
    // do nothing
   } else if(cc==byte.class)
    mw.visitInsn(ICONST_0);
   else if(cc==char.class)
    mw.visitInsn(ICONST_0);
   else if(cc==double.class)
    mw.visitInsn(DCONST_0);
   else if(cc==float.class)
    mw.visitInsn(FCONST_0);
   else if(cc==int.class)
    mw.visitInsn(ICONST_0);
   else if(cc==long.class)
    mw.visitInsn(LCONST_0);
   else if(cc==short.class)
    mw.visitInsn(ICONST_0);
   else if(cc==boolean.class)
    mw.visitInsn(ICONST_0);
   else {
    throw new InternalError(0x3C2,   
     "Internal error: unknown class in genDummyReturn()."
    );
   }
  } else {
   mw.visitInsn(ACONST_NULL);
  }
  genReturn(mw,cc);
 }
 

 private static String methodSignature(Method m) {
  return getSignature(m.getParameterTypes(),m.getReturnType());
 }

 private static String constructorSignature(Constructor c) {
  return getSignature(c.getParameterTypes(),void.class);
 }

 private static String getSignature(Class[] params,Class returnType) {
  int i;
  String s;
  int nParams=params.length;
  s="(";
  for (i=0;i<nParams;i++) {
   Class c=params[i];
   s=s+classSignature(c);
  }
  s=s+")"+classSignature(returnType);
  return s;
 }

  private static String classSignature(Class c) {
   if (c.isPrimitive()) {
    if (c==byte.class)
     return "B";
    else if(c==char.class)
     return "C";
    else if(c==double.class)
     return "D";
    else if(c==float.class)
     return "F";
    else if(c==int.class)
     return "I";
    else if(c==long.class)
     return "J";
    else if(c==short.class)
     return "S";
    else if(c==boolean.class)
     return "Z";
    else if(c==void.class)
     return "V";
    else {
     throw new InternalError(0x3C3,   
      "Internal error: unknown code in classSignature()."
     );
    }
   } else {
    // Could be an array
    Class base=c.getComponentType();
    if (base!=null)
     return "["+classSignature(base);
    else
     return "L"+c.getName().replace('.','/')+";";
   }
  }


  private static void theConst(CodeVisitor mw,int n) {
// ICONST_n for n<=5, else BIPUSH n / SIPUSH n / etc.
   switch (n) {
    case 0: mw.visitInsn(ICONST_0); break;
    case 1: mw.visitInsn(ICONST_1); break;
    case 2: mw.visitInsn(ICONST_2); break;
    case 3: mw.visitInsn(ICONST_3); break;
    case 4: mw.visitInsn(ICONST_4); break;
    case 5: mw.visitInsn(ICONST_5); break;
    default: {
     if (n<=127 && n>=-128)
      mw.visitIntInsn(BIPUSH,n);
     else if (n<=32767 && n>=-32768)
      mw.visitIntInsn(SIPUSH,n);
     else
      mw.visitLdcInsn(new Integer(n));
    }
   }

  }

  
  // factories can only be used for code that does not use special dispatch
  // much code is similar to createMasterCode, below. In fact, I could refactor a few things.
  // field initialization of the new instance is done here, rather than in the <init> of the
  // master class.
  static byte[] createFactoryCode(java.util.List<MethodGroup> groups,String nakedClassName,
    Class t,Class[] comboClasses,String[] comboMangled,List<Field> selfList) {
   
   String   className=thisPackage+"PolyD$"+nakedClassName;
   String factoryName=thisPackage+"Factory$"+nakedClassName;
   
   ClassWriter cw = new ClassWriter(true); // please, calculate maxs for me


   cw.visit(java15?V1_5:V1_1, ACC_PUBLIC | ACC_FINAL,
        factoryName,
        thisPackage+"Factory",
        new String[] {}, null);

   
   CodeVisitor mw;

   mw = cw.visitMethod(ACC_PUBLIC, "<init>",
     "()V", null, null);
   mw.visitVarInsn(ALOAD, 0);
   // super
   mw.visitMethodInsn(INVOKESPECIAL, thisPackage+"Factory", "<init>", "()V");
   mw.visitInsn(RETURN);
   mw.visitMaxs(0,0); // automatic

   
   
   // a getDispatchern is required. Up to maxFactParams bodies, I use a direct call.

   String theGetName="getDispatcher";
   String params;
   if (comboClasses.length<=PolyD.maxFactParams) {
    theGetName+=comboClasses.length;
    params="(";
    for (int i=0;i<comboClasses.length;i++)
     params+="Ljava/lang/Object;";
    params+=")Ljava/lang/Object;";
   } else {
    theGetName+="N";
    params="([Ljava/lang/Object;)Ljava/lang/Object;";
   }

   mw = cw.visitMethod(ACC_PUBLIC, theGetName,
     params, null, null);
   
   // create the new dispatcher
   mw.visitTypeInsn(NEW, className);
   mw.visitInsn(DUP);
   mw.visitMethodInsn(INVOKESPECIAL, className, "<init>", "()V");

   // initialize the combos in the new dispatchers!
   
   if (comboClasses.length<=PolyD.maxFactParams) {
    for (int i=0;i<comboClasses.length;i++) {
     mw.visitInsn(DUP); // the dispatcher
     mw.visitVarInsn(ALOAD, i+1);
     mw.visitTypeInsn(CHECKCAST, comboClasses[i].getName().replace('.','/'));
     mw.visitFieldInsn(
       PUTFIELD, className,
       comboMangled[i],"L"+comboClasses[i].getName().replace('.','/')+";");
    }
   } else {
    for (int i=0;i<comboClasses.length;i++) {
     mw.visitInsn(DUP); // the dispatcher
     mw.visitVarInsn(ALOAD, 1);
     theConst(mw,i);
     mw.visitInsn(AALOAD);
     mw.visitTypeInsn(CHECKCAST, comboClasses[i].getName().replace('.','/'));
     mw.visitFieldInsn(
       PUTFIELD, className,
       comboMangled[i],"L"+comboClasses[i].getName().replace('.','/')+";");
    }
   }

   // initialize self fields in these combos!
   
   
   for (Field f : selfList) {
    int leni = comboClasses.length;
    for (int ii=0;ii<leni;ii++) {   
     if(f.getDeclaringClass().isAssignableFrom(comboClasses[ii])) {
      mw.visitInsn(DUP); // the dispatcher
      mw.visitVarInsn(ALOAD, ii+1); // the combo
      mw.visitTypeInsn(CHECKCAST, comboClasses[ii].getName().replace('.','/'));
      mw.visitInsn(SWAP);
      mw.visitFieldInsn(
        PUTFIELD, comboClasses[ii].getName().replace('.','/'),
        f.getName(),"L"+f.getType().getName().replace('.','/')+";");
      break;
     }
    }
   }
    
   // all done
    
   mw.visitInsn(ARETURN);
   mw.visitMaxs(0,0); // automatic

   
   return cw.toByteArray();
  }


///////////////////////////////////////


  // it is assumed that comboClasses is identical to the field comboClasses in all
  // the MethodGroups in groups. If that is no longer true, the code will have to be revised.

  static byte[] createMasterCode(java.util.List<MethodGroup> groups,String nakedClassName,
    Class t,Class[] comboClasses,String[] comboMangled,boolean specialDispatch) {

   Class specialDispatchClass=null;
   if (specialDispatch) specialDispatchClass=comboClasses[0];

   String mapType=null,mapName=null;
   
   String className=thisPackage+"PolyD$"+nakedClassName;
   
   ClassWriter cw = new ClassWriter(true); // please, calculate maxs for me


/*
 * If one combo, and combo IMPLEMENTS the interface, then this master class
 * EXTENDS the combo! Else, this master class IMPLEMENTS the interface.
 */
if (specialDispatch)
   cw.visit(java15?V1_5:V1_1, ACC_PUBLIC | ACC_FINAL,
     className,
     specialDispatchClass.getName().replace('.','/'),
     new String[] {}, null);
else
   cw.visit(java15?V1_5:V1_1, ACC_PUBLIC | ACC_FINAL,
     className,
     "java/lang/Object",
     new String[] {t.getName().replace('.','/')}, null);


   CodeVisitor mw;
   
   
   // The important stuff: the routines that will call the stubs
   int index=0;
   for (MethodGroup g : groups) {



      // Are we using remapNull() ?
      // No, if the definition in Dispatching is
      // not being overridden
      boolean useRemapNull=false;
      try {
       useRemapNull=
         (g.protoInfo.theDispatching.getClass().
           getMethod("remapNull",new Class[]{Class[].class,Method[].class}).
           getDeclaringClass() != ovm.polyd.policy.Dispatching.class
          ||
          g.protoInfo.theDispatching.getClass().
           getMethod("remapNull",new Class[]{Class[].class,Method[].class,
                     Object[].class,Class[].class,Method.class,String.class}).
           getDeclaringClass() != ovm.polyd.policy.Dispatching.class
         );
      } catch(Exception e) {
       throw new InternalError(0x3C5,   
        "Internal error: remapNull() inexistent or not accessible."
       );
      }
      // If remapNull is used, we might calculate classes twice: the first
      // unhandled null argument causes an exception, and regardless of the
      // way we were loading classes (individually on the stack, or in an
      // array) the stack is cleared, so we start over creating an array
      // for remapNull. That is not particularly efficient: if we were
      // already loading classes into an array (because wee have more than
      // maxMethParams arguments) we could trap individual loads, and
      // somehow continue. But a null argument is mainly considered to be
      // a sort of error condition anyway, so top efficiency is not
      // supposed to be an issue.
      //--
      
      


    // note that for non-caching policies, even zero-arity
    // messages need to be selected at runtime using the policy:
    // it might be that sometimes the policy returns one of
    // the available methods, sometimes returns nothing

/////
// BEGIN: non-caching dispatching policy

      // if caching is disabled, delegate everything to
      // the dispatching policy

      if (g.protoInfo.theDispatching.disableCaching()) {

       g.dispatchInfoNeeded=METHODGROUP;
       mapType="MethodGroup";
       mapName="map_"+index;
       String s3=methodSignature(g.proto);
       mw = cw.visitMethod(ACC_PUBLIC|ACC_FINAL, g.proto.getName(), s3, null, null);

       Label start=new Label();
       Label end=new Label();
       Label remap=new Label();
        
       if (useRemapNull) mw.visitLabel(start);

       mw.visitFieldInsn(
         GETSTATIC, className,
         mapName,"L"+thisPackage+mapType+";");
       // fill the array of args appropriately.
       loadMapTypes(className, mw, g, true); // true -> toArray

       Class[] pars=g.proto.getParameterTypes();
       int firstSlot=firstFreeSlot(pars);
       mw.visitInsn(DUP);
       mw.visitVarInsn(ASTORE, firstSlot+1); // save the array of classes/raw args
                         // (for use by OnMissing)

       mw.visitMethodInsn(INVOKEVIRTUAL, thisPackage+mapType,
         "findBestMatch","([Ljava/lang/Object;)I");
       
       if (useRemapNull) mw.visitLabel(end);

        generateSwitchSimple(mw,g,className,comboClasses,comboMangled,mapName,mapType,
                              specialDispatch,specialDispatchClass);

       if (useRemapNull) {
        mw.visitLabel(remap);
        mw.visitTryCatchBlock(start,end,remap,"java/lang/NullPointerException");
        mw.visitInsn(POP); // discard the exception
        mw.visitFieldInsn(
         GETSTATIC, className,
         mapName,"L"+thisPackage+mapType+";");
        loadMapTypesWithRemap(className, mapName,
         mapType, mw, g);
        mw.visitJumpInsn(GOTO,end);
       }
       mw.visitMaxs(0,0); // automatic
      }

// END: non-caching dispatching policy
/////


    else


    // caching policies from here on


    // if arity zero (no args, and no raws)
    // logic would assume that a single implementing
    // method exists. Unfortunately, that is not necessarily the
    // case: duplicates might legitimately exist, and some sort of
    // selection would have to be performed. Statically, though.

    // Just call bestMatch once from here, and find the correct
    // method (if there is one)

    if (g.arityRaw==0) {
     
     mw = cw.visitMethod(ACC_PUBLIC|ACC_FINAL, g.proto.getName(), methodSignature(g.proto), null, null);
     
      int best=g.findBestMatch(new Object[]{}); // warning: g.findBestMatch return 0,1..n
      if (best==0) {
       throw new DispatcherCreationException(0x301,
        "No compatible method found for "+Utils.methodString(g.proto)
       );
      }

      best--; // move from 1..n to 0..n-1;

     if (g.protoInfo.theInvocation==null) {
      createDirectDispatch(mw, g, best, className, comboMangled,
                            specialDispatch,specialDispatchClass);
     } else {
      g.dispatchInfoNeeded=METHODGROUP;
      mapType="MethodGroup";
      mapName="map_"+index;
      createCustomDirectDispatch(mw, g, best, className, comboMangled, mapName, mapType,
                            specialDispatch,specialDispatchClass);
     }
     mw.visitMaxs(0,0); // automatic
    } else {


     // A tempting optimization would be to call directly the only
     // available body if g.bodies.length==1. BUT the only body
     // might not be compatible according to this policy, even though
     // it looks compatible. So, generate the map in every case,
     // at least until a better approach is found.


/* disabled

     for (i=0;i<g.protos.length;i++) {
      // the getParameterTypes() is also done in methodSignature,
      // later. It would be nice to avoid the duplication
      Class[] pars=g.getFiltered();
      
      int nBodies=0;
      int bodyIndex=-1;
      int nCases=g.bodies.length;
      for (int kk=0;kk<nCases;kk++) {
       if (Verifier.classArrayDiff(g.bodies[kk].getParameterTypes(),pars)!=NotCompatible) {
        if (nBodies==0) {
         bodyIndex=kk;
         nBodies++;
        } else {
         nBodies++;
         break; // >1 --> full MultiHash, ==1 --> Direct dispatch, ==0 --> internal error!!!
        }
       }
      }
      
      if (nBodies==0) {
       throw new InternalException(0x3C4,   
        "Internal error: nBodies==0."
       );
      }
      if (nBodies==1) {
       mw = cw.visitMethod(ACC_PUBLIC|ACC_FINAL, g.proto.getName(), methodSignature(g.proto), null, null);
       if (g.theInvocation==null) {
        createDirectDispatch(mw, g, bodyIndex, className, comboMangled,
                            specialDispatch,specialDispatchClass);
       } else {
        g.dispatchInfoNeeded=METHODGROUP;
        mapType="MethodGroup";
        mapName="map_"+index;
        createCustomDirectDispatch(mw, g, bodyIndex, className, comboMangled,mapName,mapType,
                            specialDispatch,specialDispatchClass);
       }
       mw.visitMaxs(0,0); // automatic
      }

// disabled until here
*/

      

      // we need to generate a dispatching routine.


      // so, we need to generate a dispatching routine. The layout of the
      // local frame will be:
      // 0 : this
      // 1..firstSlot-1 : arguments -> INCLUDING RAW ARGS
      // firstSlot: will contain the masked hash
      // firstSlot+1... will contain the classes obtained from the arguments -> WITHOUT THE RAW ARGS
      //
      // if g.arity>PolyD.maxMethParams, then firstSlot+1 contains
      // one ARRAY of classes WITH the raw args

      // note that in the list of arguments (1..firstSlot-1) there
      // could be floats or doubles, so firstSlot is not always ==arityRaw.
      // On the other hand, the set in firstSlot+1... only contain
      // Classes, so there is one slot per arg.
     
      Class[] pars=g.proto.getParameterTypes();
      int firstSlot=firstFreeSlot(pars);

      // note: g.arity does *not* includes raw arguments
      // g.arityRaw *does* include raw arguments
      
      g.dispatchInfoNeeded=MAP;
      mapType="MultiHash";
      if (g.arityRaw<=PolyD.maxMethParams)
       mapType+=g.arityRaw;
      else
       mapType+="N";
      mapName="map_"+index;
      String s3=methodSignature(g.proto);
      mw = cw.visitMethod(ACC_PUBLIC|ACC_FINAL, g.proto.getName(), s3, null, null);


      // if g.arityRaw>Multi.maxMethParams, then delegate everything to
      // the MultiHashn.
      
      if (g.arityRaw>PolyD.maxMethParams) {
       
       
       Label start=new Label();
       Label end=new Label();
       Label remap=new Label();
        
       if (useRemapNull) mw.visitLabel(start);
       
       mw.visitFieldInsn(
         GETSTATIC, className,
         mapName,"L"+thisPackage+mapType+";");

       // fill the array of args appropriately.
       loadMapTypes(className, mw, g, true); // true -> toArray

       if (useRemapNull) mw.visitLabel(end);


       // save a copy of the array of classes
       mw.visitInsn(DUP);
       mw.visitVarInsn(ASTORE, firstSlot+1); // save the array of classes/raw args
       
              mw.visitMethodInsn(INVOKEVIRTUAL, thisPackage+mapType,
         "findCode","([Ljava/lang/Object;)I");
       // ends with an integer index on the stack, used by the switch.
        generateSwitch(mw,g,className,comboClasses,comboMangled,mapName,mapType,
                            specialDispatch,specialDispatchClass);

       if (useRemapNull) {
        mw.visitLabel(remap);
        mw.visitTryCatchBlock(start,end,remap,"java/lang/NullPointerException");
        mw.visitInsn(POP); // discard the exception
        mw.visitFieldInsn(
         GETSTATIC, className,
         mapName,"L"+thisPackage+mapType+";");
        loadMapTypesWithRemap(className, mapName,
         mapType, mw, g);
        mw.visitJumpInsn(GOTO,end);
       }
      
      
      } else {


// more complicated, I include the hashcode calculations and the fast-path
// lookup directly in this code!
       
       
       Label start=new Label();
       Label end=new Label();
       Label remap=new Label();
        
       if (useRemapNull) mw.visitLabel(start);
       
             

       int slotIndex=1;
       int nonRawCounter=0;
       for (int k=0;k<g.arityRaw;k++) {
        if (k!=0) {
         theConst(mw,37);
         mw.visitInsn(IMUL);
        }
        // rather than getClass(), the following code is necessary.

        slotIndex = loadCorrectClass(g, false, className, mw, slotIndex, k, false);
                      // false -> exception on null non-raw args, false -> don't wrap ints

        // on the stack top, either an int (raw) or an Object
        
        // if an int raw arg, skip the hashCode() and do not store
        if (g.protoInfo.raw==null || g.protoInfo.raw[k]!=int.class) {
         if (g.protoInfo.raw==null || g.protoInfo.raw[k]==null) {
          mw.visitInsn(DUP);
          mw.visitVarInsn(ASTORE, firstSlot+1+nonRawCounter++); // save the class (if not raw)
         }
         // Apparently hashCode() is faster (!)
         // mw.visitMethodInsn(INVOKESTATIC, "java/lang/System", "identityHashCode", "(Ljava/lang/Object;)I");
         mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Object", "hashCode", "()I");
        }
        // accumulate
        if (k!=0) {
         mw.visitInsn(IADD);
        }
       }

       
       if (useRemapNull) {
        mw.visitLabel(end);
       }
       // if remapNull is called, execution will resume here
      
      
      
      // computation done, now mask and store:
      theConst(mw,PolyD.guessTableSize(g)-1);
      mw.visitInsn(IAND);
      mw.visitVarInsn(ISTORE,firstSlot); // masked hashcode


      // good. Now the tests
      
      Label away=new org.objectweb.asm.Label();
      Label popaway=new org.objectweb.asm.Label();
      Label doSwitch=new org.objectweb.asm.Label();




      nonRawCounter=0;
      slotIndex=1;
      for (int k=0;k<g.arityRaw;k++) {

// if int, extracts the integer from the array element,
// else compare, ptr for classes, equals for the raw args

       if (g.protoInfo.raw!=null && g.protoInfo.raw[k]==int.class) { // int
        // get the Integer from the array, and compare
         mw.visitFieldInsn(
           GETSTATIC, className,
           keysName(k,mapName),"[Ljava/lang/Object;");
        mw.visitVarInsn(ILOAD,firstSlot); // masked hashcode
        mw.visitInsn(AALOAD);
        // warning! It could be null!
        mw.visitInsn(DUP);
        mw.visitJumpInsn(IFNULL,popaway);
        unwrapResult(mw,int.class); // get the int
        mw.visitVarInsn(ILOAD,slotIndex); // retrieve the int raw arg
        mw.visitJumpInsn(IF_ICMPNE,away);
       } else {
        if (g.protoInfo.raw!=null && g.protoInfo.raw[k]!=null) { // raw arg
         mw.visitVarInsn(ALOAD,slotIndex); // retrieve the raw arg
         mw.visitFieldInsn(
           GETSTATIC, className,
           keysName(k,mapName),"[Ljava/lang/Object;");
         mw.visitVarInsn(ILOAD,firstSlot); // h
         mw.visitInsn(AALOAD);
         // it survived hashCode(), so it is not null.
         mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Object", "equals", "(Ljava/lang/Object;)Z");
         mw.visitJumpInsn(IFEQ,away); // super-stupid convention. if bool EQ <-> false (C-style)
        } else { // regular argument
         mw.visitVarInsn(ALOAD,firstSlot+1+nonRawCounter++); // retrieve the class
         mw.visitFieldInsn(
           GETSTATIC, className,
           keysName(k,mapName),"[Ljava/lang/Object;");
         mw.visitVarInsn(ILOAD,firstSlot); // h
         mw.visitInsn(AALOAD);
         mw.visitJumpInsn(IF_ACMPNE,away);
        }
       }
       slotIndex=skipItem(slotIndex, pars[k]);
      }

      // excellent. classes are the same, retrieve from the quick array.
      mw.visitFieldInsn(
        GETSTATIC, className,
        quickName(mapName),"[I");
      mw.visitVarInsn(ILOAD,firstSlot); // h
      mw.visitInsn(IALOAD);

      // ready for the switch block!


      mw.visitJumpInsn(GOTO,doSwitch);

      // away: comparison failed, so we need to go through findCode()
      
      mw.visitLabel(popaway);
      mw.visitInsn(POP); // discard one element from the stack

      mw.visitLabel(away);
      mw.visitFieldInsn(
        GETSTATIC, className,
        mapName,"L"+thisPackage+mapType+";"); // MultiMap
      String st="(I";
      mw.visitVarInsn(ILOAD,firstSlot); // h
      nonRawCounter=0;
      slotIndex=1;
      for (int k=0;k<g.arityRaw;k++) {
       if (g.protoInfo.raw==null || g.protoInfo.raw[k]==null) { // non raw arg
        mw.visitVarInsn(ALOAD,firstSlot+1+nonRawCounter++); // load class, argument to findCode
       } else {
        loadWrappedArgument(mw, slotIndex, pars[k]); // load the (raw) arg rather than the class
       }
       st+="Ljava/lang/Object;";
       slotIndex=skipItem(slotIndex, pars[k]);
      }
      st+=")I";
      mw.visitMethodInsn(INVOKEVIRTUAL, thisPackage+mapType, "findCode", st);
      
      
      mw.visitLabel(doSwitch);
      
      /*
       * begin of the switch
       */
      generateSwitch(mw,g,className,comboClasses,comboMangled,mapName,mapType,
                            specialDispatch,specialDispatchClass);
       /*
       * end of the switch
       */
      
      // if required, let's add the remapNull handler.
      // To understand what is being done, refer to the beginning
      // of this part of code, where the labels begin and end are visited
      
      if (useRemapNull) {
       mw.visitLabel(remap);
       mw.visitTryCatchBlock(start,end,remap,"java/lang/NullPointerException");
       mw.visitInsn(POP); // discard the exception
       loadMapTypesWithRemap(className, mapName, mapType, mw, g); // reload (in an array)

       // we now have on the stack a nice array of classes that we have
       // to inspect, add in various ways, and store in the appropriate
       // slots
       // we keep the array on the stack, and the accumulated result over it on top.
       
 
       nonRawCounter=0;
       slotIndex=1;
       for (int k=0;k<g.arityRaw;k++) {
        if (k!=0) {
         theConst(mw,37);
         mw.visitInsn(IMUL);
        }

        
        // if g.arity == 1, do nothing, we'll eat up the
        // array and not accumulate
        if (g.arityRaw!=1) {
         if (k==0) {
          mw.visitInsn(DUP); // only the array on top, twice
         } else {
          // initially array, and accumulator on top
          mw.visitInsn(SWAP); // array on top, accum below
          // if last element, we do not need to preserve the array
          // else push down a copy
          if (k!=g.arityRaw-1) mw.visitInsn(DUP_X1); // array on top, accum, array
         }
        }
        // get this element. IF a raw argument, OVERWRITE the original arguments
        theConst(mw,k);
        mw.visitInsn(AALOAD);
        if (g.protoInfo.raw==null || g.protoInfo.raw[k]==null) { // non raw arg
         // I have the class (bona-fide non-null)                
         mw.visitInsn(DUP); // one copy for the hashCode again
         mw.visitVarInsn(ASTORE, firstSlot+1+nonRawCounter++); // save the class
         mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Object", "hashCode", "()I");
        } else {
         // raw
         if (g.protoInfo.raw[k]==int.class) { // int
          unwrapResult(mw,int.class); // get the int
          mw.visitInsn(DUP); // one copy for the hashCode again
          mw.visitVarInsn(ISTORE, slotIndex); // save the class
         } else {
          mw.visitInsn(DUP); // one copy for the hashCode again
          mw.visitVarInsn(ASTORE, slotIndex); // save the class
          mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Object", "hashCode", "()I");
         }
        }
        // Apparently hashCode() is faster (!)
        // mw.visitMethodInsn(INVOKESTATIC, "java/lang/System", "identityHashCode", "(Ljava/lang/Object;)I");
        if (k!=0) {
         mw.visitInsn(IADD); // accumulate
        }
        slotIndex=skipItem(slotIndex, pars[k]);
       }

       mw.visitJumpInsn(GOTO,end);
      }

      
      
      }
      mw.visitMaxs(0,0); // automatic

    }
    if (g.dispatchInfoNeeded==MAP || g.dispatchInfoNeeded==METHODGROUP) index++; // keep count of used MultiHashes
   }
   
   // create the fields for the instances, always. (except for specialDispatch)
   //
if (!specialDispatch) {
   int i;
   for (i=0;i<comboClasses.length;i++) {
    cw.visitField(
      // visibility: package. Not final, could be inited by factory
     0 /* ACC_FINAL | ACC_PRIVATE*/, comboMangled[i],
     "L"+comboClasses[i].getName().replace('.','/')+";",
     null,null);
   }
}

   // if caches for classes are needed, generate them.
   if (!java15) {
    Vector<Class> done=new Vector<Class>(10,10);
//  for (MethodGroup g : groups) {
    for(java.util.Iterator ig = groups.iterator(); ig.hasNext();) {
     MethodGroup g2 = (MethodGroup)ig.next();
     prepareClassConstantCache(cw,g2.protoInfo.asClasses,className,done);
     prepareClassConstantCache(cw,g2.protoInfo.nullDefaults,className,done);
    }
   }
    
   // One map per group.
   // Together with the map, I am also creating the main arrays that will be
   // used: the int[] array and the various Object[] arrays. If the arity is
   // greater than the limit, I delegate the whole map logic to MultiHashn,
   // since incorporating the code in each dispatcher would be rather complex.
   index=0;
// for (MethodGroup g : groups) {
   java.util.Iterator done2 = groups.iterator();
   while (done2.hasNext()) {
    MethodGroup g = (MethodGroup)done2.next();

    if (g.dispatchInfoNeeded!=NONE) {
     if (g.dispatchInfoNeeded==MAP) {
      mapType="MultiHash";
      if (g.arityRaw<=PolyD.maxMethParams)
       mapType+=g.arityRaw;
      else
       mapType+="N";
      mapName="map_"+index++;
      if (g.arityRaw<=PolyD.maxMethParams) {
       cw.visitField(
         ACC_FINAL | ACC_PRIVATE | ACC_STATIC , quickName(mapName),
         "[I",
         null,null);
       if (g.arityRaw<=PolyD.maxMethParams) {
        for (int jj=0;jj<g.arityRaw;jj++) {
         cw.visitField(
           ACC_FINAL | ACC_PRIVATE | ACC_STATIC , keysName(jj,mapName),
           "[Ljava/lang/Object;",
           null,null);
        }
       }
      }
     } else if (g.dispatchInfoNeeded==METHODGROUP) {
      mapType="MethodGroup";
      mapName="map_"+index++;
     } else {
      throw new InternalError(0x3C6,   
       "Internal error: unknown map type"
      );
     }
     cw.visitField(
       ACC_FINAL | ACC_PRIVATE | ACC_STATIC , mapName,
       "L"+thisPackage+mapType+";",
       null,null);
    }
   }
   
   
   // special static initializer, MUST be called by PolyD before
   // any instances are created
   // constructor
   
   mw = cw.visitMethod(ACC_PUBLIC | ACC_STATIC | ACC_FINAL, "SPECIAL$$STATIC$$INIT",
     "(Ljava/util/List;)V", null, null);
   // Iterator iterator = list.iterator();
   mw.visitVarInsn(ALOAD, 0);
   mw.visitMethodInsn(INVOKEINTERFACE, "java/util/List", "iterator", "()Ljava/util/Iterator;");
   mw.visitVarInsn(ASTORE, 1); // iterator
   
   index=0;
// for (MethodGroup g : groups) {
   java.util.Iterator done3 = groups.iterator();
   while (done3.hasNext()) {
    MethodGroup g = (MethodGroup)done3.next();

    if (g.dispatchInfoNeeded==MAP) {
     mapType="MultiHash";
     if (g.arityRaw<=PolyD.maxMethParams)
      mapType+=g.arityRaw;
     else
      mapType+="N";
     mapName="map_"+index++;
    // map_0 = new MultiHash2(values,keys...,(MethodGroup)iterator.next());


     // first step: create the empty hashmap
     mw.visitTypeInsn(NEW, thisPackage+mapType);
     mw.visitInsn(DUP); // save a copy for the later putfield
     /*
      * initialization of the map is more complex now. If g.arity>Multi.maxMethParams,
      * the raw arrays are created, and used as parameters for initialization of the
      * MultiHash.
      */
     if (g.arityRaw<=PolyD.maxMethParams) {
      int size=PolyD.guessTableSize(g);
      // quick array = new int[size];
      theConst(mw,size);
      mw.visitIntInsn(NEWARRAY,T_INT);
      mw.visitInsn(DUP); // save a copy for later <init> 
      mw.visitFieldInsn(
        PUTSTATIC, className,
        quickName(mapName),"[I");
      // keyx arrays = new Class[size];
      for (int ii=0;ii<g.arityRaw;ii++) {
       theConst(mw,size);
       mw.visitTypeInsn(ANEWARRAY, "java/lang/Object");
       mw.visitInsn(DUP); // save a copy for later <init>     
       mw.visitFieldInsn(
        PUTSTATIC, className,
        keysName(ii,mapName),"[Ljava/lang/Object;");     
     }

     } // else do nothing. (array of classes, MultiHashn will take care of it)
     //
     mw.visitVarInsn(ALOAD, 1);
     mw.visitMethodInsn(INVOKEINTERFACE, "java/util/Iterator", "next", "()Ljava/lang/Object;");
     mw.visitTypeInsn(CHECKCAST, thisPackage+"MethodGroup");
     // ready to call <init>. Signature goes like:
     // if g.arityRaw>Multi.maxMethParams then "(L"+thisPackage+"MethodGroup;)V", 
     // else ([I plus [Ljava/lang/Class; g.arity times, plus L.../MethodGroup;)V.
     String initSignature="(";
     if (g.arityRaw<=PolyD.maxMethParams) {
      initSignature+="[I";
      for (int ii=0;ii<g.arityRaw;ii++) {
       initSignature+="[Ljava/lang/Object;";
      }
     }
     initSignature+="L"+thisPackage+"MethodGroup;)V";
     mw.visitMethodInsn(INVOKESPECIAL, thisPackage+mapType, "<init>", initSignature);
     mw.visitFieldInsn(
       PUTSTATIC, className,
       mapName,"L"+thisPackage+mapType+";");
    } // end case g.dispatchInfoNeeded==MAP
    else if (g.dispatchInfoNeeded==METHODGROUP) {
     // initialize the instance variable with the MethodGroup
     mapType="MethodGroup";
     mapName="map_"+index++;
     mw.visitVarInsn(ALOAD, 1); // retrieve the iterator
     mw.visitMethodInsn(INVOKEINTERFACE, "java/util/Iterator", "next", "()Ljava/lang/Object;");
     mw.visitTypeInsn(CHECKCAST, thisPackage+"MethodGroup");
     mw.visitFieldInsn(
       PUTSTATIC, className,
       mapName,"L"+thisPackage+mapType+";");
    } else { // g.dispatchInfoNeeded==NONE
     // if the MethodGroup represents a message with no parameters, then it must
     // be skipped. That could be done before passing the list to this constructor,
     // but the code has to executed anyway, so doing it here is more or less
     // the same.
     mw.visitVarInsn(ALOAD, 1);
     mw.visitMethodInsn(INVOKEINTERFACE, "java/util/Iterator", "next", "()Ljava/lang/Object;");
     mw.visitInsn(POP);
    }
   }
   mw.visitInsn(RETURN);
   mw.visitMaxs(0,0); // automatic
   


   // the handling of initialization is quite different for
   // regular dispatchers and for specialDispatch dispatchers.
   
   // in the first case, we only have <init>(Object[])
   
   // in the second case, all constructors of the specialDispatchClass
   // are replicated, doing nothing but wrapping the constructor of
   // the super. Only public constructors are wrapped. Exceptions
   // thrown by super constructors are redeclared.
   
   
   if (!specialDispatch) {

   // constructor does not actually do anything, initialization
   // is done by the Factory
   mw = cw.visitMethod(ACC_PUBLIC, "<init>",
     "()V", null, null);
   // super
   mw.visitVarInsn(ALOAD, 0);
   mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>", "()V");
   mw.visitInsn(RETURN);
   mw.visitMaxs(0,0); // automatic

   } else {
    
    Constructor[] cs=specialDispatchClass.getConstructors();
    
    int csl=cs.length;
    for (int csi=0;csi<csl;csi++) {
     Constructor c=cs[csi];
     String sig=constructorSignature(c);
     Class [] ecs=c.getExceptionTypes();
     String [] ess=new String[ecs.length];
     for (int eci=0;eci<ecs.length;eci++)
      ess[eci]=classSignature(ecs[eci]);
     mw = cw.visitMethod(ACC_PUBLIC, "<init>",
       sig, ess, null);

     mw.visitVarInsn(ALOAD, 0);

     // load all arguments.
     Class[] params=c.getParameterTypes();
     int arity=params.length;
     int slotIndex=1;
     for (int k=0;k<arity;k++) {
       Class cc=params[k];
       slotIndex=loadArgument(mw, slotIndex, cc);
     }

    //
     mw.visitMethodInsn(INVOKESPECIAL, specialDispatchClass.getName().replace('.','/'),
       "<init>", sig);
     mw.visitInsn(RETURN);
     mw.visitMaxs(0,0); // automatic
    
    }
   }


   return cw.toByteArray();
  }
//**************************************************************
//**************************************************************

  
  /* 
   * loads the list of classes required for the MultiHash lookup.
   * If "toArray" is true, load the classes and the raw arguments
   * into an Object[], otherwise load them individually on the stack.
   *
   * IF toarray, then wrap raw integer args
   */
  private static void loadMapTypes(String className, CodeVisitor mw, MethodGroup g, boolean toArray) {
   int slotIndex=1;
   if (toArray) {
    theConst(mw,g.arityRaw); // elements to be moved into the array
    mw.visitTypeInsn(ANEWARRAY, "java/lang/Object");
   }
   for (int k=0;k<g.arityRaw;k++) {
    if (toArray) {
     mw.visitInsn(DUP);
     theConst(mw,k);
    }
    slotIndex = loadCorrectClass(g, false, className, mw, slotIndex, k,toArray); // false -> exception on null non-raw args
    if (toArray) mw.visitInsn(AASTORE);
   }
  }

  
  // if loadNulls is true, EVERY argument is checked for nullness, and if it is null,
  // and a nullDefault is not provided, a null is loaded. if loadNulls is false,
  // and a null argument is dynamically found, the test is not present and a
  // nullPointerException is generated instead.
  // IF wrapRawInts, then wrap into Integers raw ints after loading them
  private static int loadCorrectClass(MethodGroup g,
        boolean loadNulls, String className,
        CodeVisitor mw, int slotIndex, int k,boolean wrapRawInts) {
   Class[] forcedClasses=g.protoInfo.asClasses;
   Class[] nullDefaults=g.protoInfo.nullDefaults;
   Class[] pars=g.proto.getParameterTypes();
   boolean raw=(g.protoInfo.raw!=null && g.protoInfo.raw[k]!=null);
   if (forcedClasses!=null && forcedClasses[k]!=null) {
    // we know it is not a raw argument if @As was used.
    loadClassConstant(mw, forcedClasses[k],className);
    slotIndex=skipItem(slotIndex, pars[k]);
   } else {
    if (g.protoInfo.raw==null || g.protoInfo.raw[k]==null)
     slotIndex=loadItem(mw, slotIndex, pars[k],className,nullDefaults!=null?nullDefaults[k]:null,loadNulls);
    else if (wrapRawInts || g.protoInfo.raw[k]!=int.class) {
     slotIndex=loadWrappedArgument(mw, slotIndex, pars[k]); // load item rather than class
    } else {
     // load int without wrapping
     mw.visitVarInsn(ILOAD, slotIndex++);
    }
   }
   return slotIndex;
  }

  // as loadMapTypes, but detects null arguments, loads everything into an array,
  // calls remapNull, leaves everything into an array
  // (client will unroll if required)
  private static void loadMapTypesWithRemap(String className,
    String mapName,String mapClass, CodeVisitor mw, MethodGroup g) {

   // time to call remapNull()
   // let's prepare the methodgroup
   mw.visitFieldInsn(
     GETSTATIC, className,
     mapName,"L"+thisPackage+mapClass+";");
   mw.visitFieldInsn(
     GETFIELD, thisPackage+mapClass,
     "methGroup","L"+thisPackage+"MethodGroup;");

   int slotIndex=1;
   theConst(mw,g.arityRaw); // elements to be moved into the array
   mw.visitTypeInsn(ANEWARRAY, "java/lang/Object");
   for (int k=0;k<g.arityRaw;k++) {
    mw.visitInsn(DUP);
    theConst(mw,k);
    slotIndex = loadCorrectClass(g, true, className, mw, slotIndex, k, true); // second true -> wrap
    mw.visitInsn(AASTORE);
   }
   
   // let's call remapNull()
   mw.visitMethodInsn(INVOKEVIRTUAL, thisPackage+"MethodGroup", "remapNull",
     "([Ljava/lang/Object;)[Ljava/lang/Object;");
   // we have the array return as it is, caller will unroll
  }

  private static void loadClassConstant(CodeVisitor mw, Class cc,String className) {
   if (java15)
    mw.visitLdcInsn(org.objectweb.asm.Type.getType(cc));
   else
    loadClassConstantCaching(mw,cc,className);
  }

  // compatibility with pre-5.0 Java - The result would *really* have to be cached.
  // There needs to be one cache for each position used in forcedClasses.
  private static void loadClassConstantNonCaching(CodeVisitor mw, Class cc,String className) {
   mw.visitLdcInsn(cc.getName());
   mw.visitMethodInsn(INVOKESTATIC, "java/lang/Class", "forName", "(Ljava/lang/String;)Ljava/lang/Class;");
  }
  
  // compatibility with pre-5.0 Java - As above, but cached.
  private static void loadClassConstantCaching(CodeVisitor mw, Class cc,String className) {
   Label here=new org.objectweb.asm.Label();
   mw.visitFieldInsn(
      GETSTATIC, className, // this class
      "classCache$"+cc.getName().replace('.','$'),"Ljava/lang/Class;");
   mw.visitInsn(DUP);
   mw.visitJumpInsn(IFNONNULL,here);
   mw.visitInsn(POP);
   mw.visitLdcInsn(cc.getName());   
   mw.visitMethodInsn(INVOKESTATIC, "java/lang/Class", "forName", "(Ljava/lang/String;)Ljava/lang/Class;");
   mw.visitInsn(DUP);
   mw.visitFieldInsn(
      PUTSTATIC, className,
      "classCache$"+cc.getName().replace('.','$'),"Ljava/lang/Class;");
   mw.visitLabel(here);
  }

  private static void prepareClassConstantCache(ClassWriter cw, Class[] forcedClasses,String className,Vector<Class> done) {
   // must not create duplicates. Since the class array is supposed to be very short,
   // I can do it wastefully. Also, be careful, some elements might be null.
   if (java15 || forcedClasses==null) return;
   
   int len=forcedClasses.length;

   here:
   for (int i=0;i<len;i++) {
    if (forcedClasses[i]!=null&&!done.contains(forcedClasses[i])) {
     done.addElement(forcedClasses[i]);
     cw.visitField(
       ACC_PRIVATE | ACC_STATIC, "classCache$"+forcedClasses[i].getName().replace('.','$'),
       "Ljava/lang/Class;",
       null,null); 
    }
   }
  }
  
/*
 * Example of caching (a simple Something.class would expand to
 * all of the following code, before Java 1.5!):
 * The code that obtaines the Class instance is:
   0:   aload_0
   1:   getstatic       #7; //Field class$java$lang$String:Ljava/lang/Class;
   4:   ifnonnull       19
   7:   ldc     #8; //String java.lang.String
   9:   invokestatic    #9; //Method class$:(Ljava/lang/String;)Ljava/lang/Class;
   12:  dup
   13:  putstatic       #7; //Field class$java$lang$String:Ljava/lang/Class;
   16:  goto    22
   19:  getstatic       #7; //Field class$java$lang$String:Ljava/lang/Class;

and the caching routine is:
static java.lang.Class class$(java.lang.String);
  Code:
   0:   aload_0
   1:   invokestatic    #1; //Method java/lang/Class.forName:(Ljava/lang/String;)Ljava/lang/Class;
   4:   areturn
   5:   astore_1
   6:   new     #3; //class NoClassDefFoundError
   9:   dup
   10:  aload_1
   11:  invokevirtual   #4; //Method java/lang/ClassNotFoundException.getMessage:()Ljava/lang/String;
   14:  invokespecial   #5; //Method java/lang/NoClassDefFoundError."<init>":(Ljava/lang/String;)V
   17:  athrow
  Exception table:
   from   to  target type
     0     4     5   Class java/lang/ClassNotFoundException

Where there's obviously a field:
static java.lang.Class class$java$lang$String;

 * I could simplify the caching assuming exceptions are never thrown, so only the first three
 * lines of the above routine would be used, and that could be inlined as follows:

   0:   aload_0
   1:   getstatic       #7; //Field class$java$lang$String:Ljava/lang/Class;
   4:   ifnonnull       19
   7:   ldc     #8; //String java.lang.String
   9:   invokestatic    #1; //Method java/lang/Class.forName:(Ljava/lang/String;)Ljava/lang/Class;
   12:  dup
   13:  putstatic       #7; //Field class$java$lang$String:Ljava/lang/Class;
   16:  goto    22
   19:  getstatic       #7; //Field class$java$lang$String:Ljava/lang/Class;

Which is a bit ugly. I change it to:

   0:   aload_0
   1:   getstatic       #7; //Field class$java$lang$String:Ljava/lang/Class;
        dup
   4:   ifnonnull       16
        pop
   7:   ldc     #8; //String java.lang.String
   9:   invokestatic    #1; //Method java/lang/Class.forName:(Ljava/lang/String;)Ljava/lang/Class;
   12:  dup
   13:  putstatic       #7; //Field class$java$lang$String:Ljava/lang/Class;
   16:  
   
   In this way I save a getstatic in the fast path. Not that it matters, since the
   code above ends up mangled and compiled in various ways, but still.
   
 * 
 * 
 */

  
  // if loadNulls is true, EVERY argument is checked for nullness, and if it is null,
  // and a nullDefault is not provided, a null is loaded. if loadNulls is false,
  // and a null argument is dynamically found, the test is not present and a
  // nullPointerException is generated instead.
  private static int loadItem(CodeVisitor mw, int slotIndex, Class cc,
     String className,Class nullDefault, boolean loadNulls) {
   if (cc.isPrimitive()) {
    String type=null;
    if (cc==byte.class)
     type="java/lang/Byte";
    else if(cc==char.class)
     type="java/lang/Character";
    else if(cc==double.class)
    {type="java/lang/Double";slotIndex++;}
    else if(cc==float.class)
     type="java/lang/Float";
    else if(cc==int.class)
     type="java/lang/Integer";
    else if(cc==long.class)
    {type="java/lang/Long";slotIndex++;}
    else if(cc==short.class)
     type="java/lang/Short";
    else if(cc==boolean.class)
     type="java/lang/Boolean";
    else {
     throw new InternalError(0x3C7,   
      "Internal error: unknown class in loadItem()."
     );
    }
    mw.visitFieldInsn(
     GETSTATIC, type,
     "TYPE","Ljava/lang/Class;");
   // Note: this is obviously a kludge, all final/primitive classes should not
   // be hashed at all.
   } else {

// scenarios:
// nullDefault:false, loadNulls:false
// -> get the class of arg. If getClass() fails, tough.
// nullDefault:false, loadNulls:true
// -> get the class of arg. If an arg is null, load null instead
// nullDefault:true
// -> get the class of arg. If arg is null, load nullDefault instead

    mw.visitVarInsn(ALOAD, slotIndex);
    if (nullDefault==null) {
     if (loadNulls==false) {
       mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Object",
        "getClass","()Ljava/lang/Class;");
     } else {
      Label loadNull=new Label();
      Label cont=new Label();

      mw.visitJumpInsn(IFNULL,loadNull);
      mw.visitVarInsn(ALOAD, slotIndex);
      mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Object",
        "getClass","()Ljava/lang/Class;");
      mw.visitJumpInsn(GOTO,cont);
      mw.visitLabel(loadNull);
      mw.visitInsn(ACONST_NULL);
      mw.visitLabel(cont);     
     }
    } else {
     // a try/catch should be more efficient, in theory,
     // but a NullPointerException wipes out the stack,
     // so it is more difficult to generate code that is
     // truly efficient using exceptions.
     Label loadDefault=new Label();
     Label cont=new Label();
     
     mw.visitJumpInsn(IFNULL,loadDefault);
     mw.visitVarInsn(ALOAD, slotIndex);
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Object",
       "getClass","()Ljava/lang/Class;");
     mw.visitJumpInsn(GOTO,cont);
     mw.visitLabel(loadDefault);
     loadClassConstant(mw, nullDefault,className);
     mw.visitLabel(cont);
    }
   }
   return ++slotIndex;
  }

  private static int skipItem(int slotIndex, Class cc) {
   if (cc==double.class || cc==long.class)
    return slotIndex+2;
   else
    return slotIndex+1;
  }
  
  static void writeCode(byte[] code,String name){
   try {
    System.out.println("Creating "+"AUTO_"+name+".class");
    FileOutputStream fos = new FileOutputStream("AUTO_"+name+".class");
    fos.write(code);
    fos.close();
   } catch(Exception e) {
    throw new InternalError(0x3C8,   
     "Internal error: could not write class file in writeCode()"
/*#ifdef14
#else14 */
      ,e
//#endif14
    );
   }
  }
  
// name of int[] array associated with the map
  private static String quickName(String s2) { return "quick_"+s2; }
  
// name of the keys[] arrays associated with the map
  private static String keysName(int jj,String s2) { return "keys"+jj+"_"+s2;}
  
  
  // The layout of the local frame will be:
  // 0 : this
  // 1..firstSlot-1 : arguments
  // firstSlot: will contain the first local var
  private static int firstFreeSlot(Class pars[]) {
   int slotIndex=1;
   for (int k=0;k<pars.length;k++) {
    slotIndex=skipItem(slotIndex, pars[k]);
   }
   return slotIndex;
  }
  
  // returns the class, already converted.
  private static String findClassAccess(CodeVisitor mw,String className,
                         Method mt,Class[] comboClasses,String comboMangled[]) {
   String instanceClass=null;
   for (int i=0;i<comboClasses.length;i++) {
    if (comboClasses[i]==mt.getDeclaringClass()) {
     instanceClass=comboClasses[i].getName().replace('.','/');
     break;
    }
   }
   return instanceClass;
  }

 // return the class, already converted.
  private static String generateInstanceAccess(CodeVisitor mw,String className,
                          Method mt,Class[] comboClasses,String comboMangled[]) {
   String instance=null;
   String instanceClass=null;
   for (int i=0;i<comboClasses.length;i++) {
    if (mt.getDeclaringClass().isAssignableFrom(comboClasses[i])) {
     instance=comboMangled[i];
     instanceClass=comboClasses[i].getName().replace('.','/');
     break;
    }
   }
   mw.visitVarInsn(ALOAD, 0);

   mw.visitFieldInsn(
     GETFIELD, className,
     instance,"L"+instanceClass+";");
   return instanceClass;
  }

  
  private static void generateSwitch(CodeVisitor mw,MethodGroup g,
    String className,Class[] comboClasses,String[] comboMangled,String mapName,String mapClass,
    boolean specialDispatch,Class specialDispatchClass) {
   generateSwitchGeneral(mw,g,className,comboClasses,comboMangled,mapName,mapClass,
                         specialDispatch,specialDispatchClass,false);
  }

  private static void generateSwitchSimple(CodeVisitor mw,MethodGroup g,
    String className,Class[] comboClasses,String[] comboMangled,String mapName,String mapClass,
    boolean specialDispatch,Class specialDispatchClass) {
   generateSwitchGeneral(mw,g,className,comboClasses,comboMangled,mapName,mapClass,
                         specialDispatch,specialDispatchClass,true);
  }

  private static void generateSwitchGeneral(CodeVisitor mw,MethodGroup g,
    String className,Class[] comboClasses,String[] comboMangled,String mapName,String mapClass,
    boolean specialDispatch,Class specialDispatchClass,boolean simple) {

   // if simple, use arrays and do not attempt to use lookupAndUpdate
   
   // if there are n bodies, there are n+1 possible alternatives.
   // The first one is "not found", then the n possible methods. Not all of
   // them, however, will ever be reachable since the corresponding body
   // might be unreachable through the current prototype (g.proto)
   // there is no point in generating code for the others. In order to
   // avoid Java complaints, the case 0 is used both to generate an
   // error (and return a fictitious return value if necessary) and as a
   // default. Note that, if we are here, there is at least one compatible body.
   
   Label defaultLabel=new Label();
   Label retrySwitchLabel=new Label();
   
   int nCases=g.bodies.length;
   Label[] targets=new Label[nCases+1];
   targets[0]=defaultLabel;

   boolean[] reachable=new boolean[nCases];
   Class[] filtParams=g.getFiltered();
   for (int kk=0;kk<nCases;kk++) {
    Diff d=Utils.classArrayDiff(g.bodies[kk].getParameterTypes(),filtParams);
    if (d==NotCompatible) {
     targets[kk+1]=defaultLabel;
    } else {
     targets[kk+1]=new Label();
     reachable[kk]=true;
    }
   }

   Class[] params=g.proto.getParameterTypes();
   int firstSlot=firstFreeSlot(params);

   if (!simple) {
   
   // I have the value for the switch on the stack.

   mw.visitInsn(DUP);
   mw.visitJumpInsn(IFNE,retrySwitchLabel);
   mw.visitInsn(POP);

   // if here, try again

   // something went wrong, index was zero.
   // try again, this time in a synchronized section
   // from here
   mw.visitFieldInsn(
     GETSTATIC, className,
     mapName,"L"+thisPackage+mapClass+";");
   String st;
   if (g.arityRaw>PolyD.maxMethParams) {
    // Retrieve the Class[]
    mw.visitVarInsn(ALOAD, firstSlot+1); // get the class array
    st="([Ljava/lang/Object;)I";
   } else {
    // in this case, if remapNull has done its thing the new vaules are already
    // available in the slots, and similarly if it has not been used.
    st="(I";
    mw.visitVarInsn(ILOAD,firstSlot); // 
    int nonRawCounter=0;
    int slotIndex=1;
    for (int k=0;k<g.arityRaw;k++) {
     if (g.protoInfo.raw==null || g.protoInfo.raw[k]==null) { // non raw arg
      mw.visitVarInsn(ALOAD,firstSlot+1+nonRawCounter++); // load class, argument to findCode
     } else {
      loadWrappedArgument(mw, slotIndex, params[k]); // load the (raw) arg rather than the class
     }
     st+="Ljava/lang/Object;";
     slotIndex=skipItem(slotIndex, params[k]);
    }
    st+=")I";
   }
   mw.visitMethodInsn(INVOKEVIRTUAL, thisPackage+mapClass,
     "lookupAndUpdate", st);
   // to here.

   mw.visitLabel(retrySwitchLabel);

   }
   mw.visitTableSwitchInsn(0,nCases,defaultLabel,targets);
   mw.visitLabel(defaultLabel);
   
   if (g.protoInfo.theMissing==Missing.ABORT) {
     mw.visitFieldInsn(
       GETSTATIC, "java/lang/System",
       "err","Ljava/io/PrintStream;");
     mw.visitLdcInsn("Fatal Error - Selector didn't find a suitable method.\nAborting.");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/io/PrintStream", "println", "(Ljava/lang/String;)V");
     theConst(mw,10); // error code
     mw.visitMethodInsn(INVOKESTATIC, "java/lang/System", "exit", "(I)V");
   } else
   if (g.protoInfo.theMissing==Missing.WARN) {
    mw.visitFieldInsn(
      GETSTATIC, "java/lang/System",
      "err","Ljava/io/PrintStream;");
    mw.visitLdcInsn("Warning - Selector didn't find a suitable method.");
    mw.visitMethodInsn(INVOKEVIRTUAL, "java/io/PrintStream", "println", "(Ljava/lang/String;)V");
   } else
   if (g.protoInfo.theMissing==Missing.FAIL) {
    mw.visitTypeInsn(NEW, thisPackage+"MissingMethodException");
    mw.visitInsn(DUP);
    mw.visitMethodInsn(INVOKESPECIAL, thisPackage+"MissingMethodException", "<init>", "()V");
    mw.visitInsn(ATHROW);
   } else
   if (g.protoInfo.theMissing==Missing.STANDARD) {
    
    // let's prepare the methodgroup
    mw.visitFieldInsn(
      GETSTATIC, className,
      mapName,"L"+thisPackage+mapClass+";");
    mw.visitFieldInsn(
      GETFIELD, thisPackage+mapClass,
      "methGroup","L"+thisPackage+"MethodGroup;");

    // rebuild an array of classes for onMissing()
    if (simple || g.arityRaw>PolyD.maxMethParams) {
     // Retrieve the Class[]
     mw.visitVarInsn(ALOAD, firstSlot+1); // get the class array
    } else {
     // must build a new array, and load the classes
     theConst(mw,g.arityRaw); // elements to be moved into the array
     mw.visitTypeInsn(ANEWARRAY, "java/lang/Object");
     int nonRawCounter=0;
     int slotIndex=1;
     for (int k=0;k<g.arityRaw;k++) {
      mw.visitInsn(DUP);
      theConst(mw,k);
      if (g.protoInfo.raw==null || g.protoInfo.raw[k]==null) { // non raw arg
       mw.visitVarInsn(ALOAD,firstSlot+1+nonRawCounter++); // load class, argument to findCode
      } else {
       loadWrappedArgument(mw, slotIndex, params[k]); // load the (raw) arg rather than the class
      }
      mw.visitInsn(AASTORE);
      slotIndex=skipItem(slotIndex, params[k]);
     }
    }
    
    // let's call remapNull()
    mw.visitMethodInsn(INVOKEVIRTUAL, thisPackage+"MethodGroup", "onMissing",
      "([Ljava/lang/Object;)V");

    // end STANDARD
    
   } else
   if (g.protoInfo.theMissing==Missing.IGNORE) {
    // do nothing
   } else {
    throw new InternalError(0x3C9,   
     "Internal error: unknown Missing Method handling."
    );
   }
   // generate a return instruction/value of some sort.
   genDummyReturn(mw,g.proto.getReturnType());
   
   
   for (int bodyIndex=0;bodyIndex<nCases;bodyIndex++) {
    if (reachable[bodyIndex]) {
     mw.visitLabel(targets[bodyIndex+1]);
//////////////
     if (g.protoInfo.theInvocation==null) {
      createDirectDispatch(mw,g,bodyIndex,className,comboMangled,
                              specialDispatch,specialDispatchClass);
     } else {
      createCustomDirectDispatch(mw,g,bodyIndex,className,comboMangled,mapName,mapClass,
                              specialDispatch,specialDispatchClass);
     }

     //
     
    }
/////////////////////////////////
   }
  }


 // this variant is used when a custom invocation policy is used.
  
  private static void createCustomDirectDispatch(CodeVisitor mw, MethodGroup g,
    int bodyIndex, String className, String[] comboMangled,
    String mapName, String mapClass, boolean specialDispatch,Class specialDispatchClass) {

   Method proto=g.proto;
   Method mt=g.bodies[bodyIndex];

   // The instance is passed in any case, even if the final method is static
   mw.visitFieldInsn(
     GETSTATIC, className,
     mapName,"L"+thisPackage+mapClass+";");
    // if edu.purdue.cs.polyd.MethodGroup then the "map" variable really refers
    // directly to a methodgroup rather than to a map
   if (mapClass!="MethodGroup") {
    mw.visitFieldInsn(
     GETFIELD, thisPackage+mapClass,
     "methGroup","L"+thisPackage+"MethodGroup;");
   }
   // MethodGroup on top
   mw.visitInsn(DUP);
   mw.visitFieldInsn(
     GETFIELD, thisPackage+"MethodGroup",
     "protoInfo","L"+thisPackage+"Descriptor$MethInfo;");
   mw.visitFieldInsn(
     GETFIELD, thisPackage+"Descriptor$MethInfo",
     "theInvocation","L"+thisPackage+"policy/Invocation;");
   mw.visitInsn(SWAP);
   // MethodGroup on top, invocation policy below
   mw.visitFieldInsn(
     GETFIELD, thisPackage+"MethodGroup",
     "bodies","[Ljava/lang/reflect/Method;");
   theConst(mw,bodyIndex);
   mw.visitInsn(AALOAD);
   // Method on top, invocation policy below
   String instanceClass;

if (specialDispatch)
   mw.visitVarInsn(ALOAD, 0);
else
   instanceClass=generateInstanceAccess(mw,className,mt,g.comboClasses,comboMangled);

   mw.visitInsn(SWAP);
   // Method on top, combo below, invocation policy below   
   
   loadArgumentsAsArray(mw, g, bodyIndex);
   // Arguments on top, then method, then combo, then invocation policy
   
   mw.visitMethodInsn(INVOKEVIRTUAL, thisPackage+"policy/Invocation",
     "invoke","(Ljava/lang/Object;Ljava/lang/reflect/Method;[Ljava/lang/Object;)Ljava/lang/Object;");

   unwrapResult(mw,proto.getReturnType());
   genReturn(mw,mt.getReturnType());
  }


  
  private static void unwrapResult(CodeVisitor mw, Class result) {
   if (result.isPrimitive()) {
    if (result==byte.class)
    {
     mw.visitTypeInsn(CHECKCAST,"java/lang/Byte");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Byte", "byteValue", "()B");
    }
    else if(result==char.class)
    {
     mw.visitTypeInsn(CHECKCAST,"java/lang/Character");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Character", "charValue", "()C");
    }
    else if(result==double.class)
    {
     mw.visitTypeInsn(CHECKCAST,"java/lang/Double");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Double", "doubleValue", "()D");
    }
    else if(result==float.class)
    {
     mw.visitTypeInsn(CHECKCAST,"java/lang/Float");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Float", "floatValue", "()F");
    }
    else if(result==int.class)
    {
     mw.visitTypeInsn(CHECKCAST,"java/lang/Integer");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Integer", "intValue", "()I");
    }
    else if(result==long.class)
    {
     mw.visitTypeInsn(CHECKCAST,"java/lang/Long");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Long", "longValue", "()J");
    }
    else if(result==short.class)
    {
     mw.visitTypeInsn(CHECKCAST,"java/lang/Short");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Short", "shortValue", "()S");
    }
    else if(result==boolean.class)
    {
     mw.visitTypeInsn(CHECKCAST,"java/lang/Boolean");
     mw.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Boolean", "booleanValue", "()Z");
    }
    else if(result==void.class)
     mw.visitInsn(POP);
    else {
     throw new InternalError(0x3CA,   
      "Internal error: unknown class in unwrapResult()."
     );
    }
   } else {
    if (!result.equals(Object.class)) {
     mw.visitTypeInsn(CHECKCAST,result.getName().replace('.','/'));
    }
   }
  }
  

  // must skip raw arguments!
  private static void loadArgumentsAsArray(CodeVisitor mw, MethodGroup g, int bodyIndex) {
   Class[] params=g.proto.getParameterTypes();
   Class[] args=g.bodies[bodyIndex].getParameterTypes();
   int arity=params.length;
   int slotIndex=1;
   int loading=0;
   theConst(mw,args.length); //<arity, in general
   mw.visitTypeInsn(ANEWARRAY, "java/lang/Object");
   for (int k=0;k<arity;k++) {
    if (g.protoInfo.raw==null || g.protoInfo.raw[k]==null) {
     mw.visitInsn(DUP);
     theConst(mw,loading++);
     slotIndex=loadWrappedArgument(mw, slotIndex, params[k]);
     mw.visitInsn(AASTORE);
    } else slotIndex=skipItem(slotIndex, params[k]);
   }  
  }
  
  private static int loadWrappedArgument(CodeVisitor mw, int slotindex, Class cc) {
   if (cc.isPrimitive()) {
    if (cc==byte.class)
    {
     if (java15) {
      mw.visitVarInsn(ILOAD, slotindex++);
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Byte", "valueOf", "(B)Ljava/lang/Byte;");
     } else {
      mw.visitTypeInsn(NEW, "java/lang/Byte");
      mw.visitInsn(DUP);
      mw.visitVarInsn(ILOAD, slotindex++);
      mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Byte", "<init>", "(B)V");
     }
    }
    else if(cc==char.class)
    {
     if (java15) {
      mw.visitVarInsn(ILOAD, slotindex++);      
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Character", "valueOf", "(C)Ljava/lang/Character;");
     } else {
      mw.visitTypeInsn(NEW, "java/lang/Character");
      mw.visitInsn(DUP);
      mw.visitVarInsn(ILOAD, slotindex++);      
      mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Character", "<init>", "(C)V");
     }
    }
    else if(cc==double.class)
    {
     if (java15) {
      mw.visitVarInsn(DLOAD, slotindex++); slotindex++; 
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Double", "valueOf", "(D)Ljava/lang/Double;");
     } else {
      mw.visitTypeInsn(NEW, "java/lang/Double");
      mw.visitInsn(DUP);
      mw.visitVarInsn(DLOAD, slotindex++); slotindex++; 
      mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Double", "<init>", "(D)V");
     }
    }
    else if(cc==float.class)
    {
     if (java15) {
      mw.visitVarInsn(FLOAD, slotindex++);      
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Float", "valueOf", "(F)Ljava/lang/Float;");
     } else {
      mw.visitTypeInsn(NEW, "java/lang/Float");
      mw.visitInsn(DUP);
      mw.visitVarInsn(FLOAD, slotindex++);      
      mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Float", "<init>", "(F)V");
     }
    }
    else if(cc==int.class)
    {
     if (java15) {
      mw.visitVarInsn(ILOAD, slotindex++);
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Integer", "valueOf", "(I)Ljava/lang/Integer;");
     } else {
      mw.visitTypeInsn(NEW, "java/lang/Integer");
      mw.visitInsn(DUP);
      mw.visitVarInsn(ILOAD, slotindex++);
      mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Integer", "<init>", "(I)V");
     }
    }
    else if(cc==long.class)
    {
     if (java15) {
      mw.visitVarInsn(LLOAD, slotindex++); slotindex++;
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Long", "valueOf", "(J)Ljava/lang/Long;");
     } else {
      mw.visitTypeInsn(NEW, "java/lang/Long");
      mw.visitInsn(DUP);
      mw.visitVarInsn(LLOAD, slotindex++); slotindex++;
      mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Long", "<init>", "(J)V");
     }
    }
    else if(cc==short.class)
    {
     if (java15) {
      mw.visitVarInsn(ILOAD, slotindex++);      
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Short", "valueOf", "(S)Ljava/lang/Short;");
     } else {
      mw.visitTypeInsn(NEW, "java/lang/Short");
      mw.visitInsn(DUP);
      mw.visitVarInsn(ILOAD, slotindex++);      
      mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Short", "<init>", "(S)V");
     }
    }
    else if(cc==boolean.class)
    {
     if (java15) {
      mw.visitVarInsn(ILOAD, slotindex++);      
      mw.visitMethodInsn(INVOKESTATIC, "java/lang/Boolean", "valueOf", "(Z)Ljava/lang/Boolean;");
     } else {
      mw.visitTypeInsn(NEW, "java/lang/Boolean");
      mw.visitInsn(DUP);
      mw.visitVarInsn(ILOAD, slotindex++);      
      mw.visitMethodInsn(INVOKESPECIAL, "java/lang/Boolean", "<init>", "(Z)V");
     }
    }
    else {
     throw new InternalError(0x3CB,   
      "Internal error: unknown class in loadWrappedArgument()"
     );
    }
   } else {
    mw.visitVarInsn(ALOAD, slotindex++);
   }
   return slotindex;
  }
  

  // creates the snippet of code necessary to directly call the correct body.
  private static void createDirectDispatch(CodeVisitor mw, MethodGroup g,
    int bodyIndex, String className, String[] comboMangled,
    boolean specialDispatch, Class specialDispatchClass) {
   
   Method mt=g.bodies[bodyIndex];
   boolean statik=(mt.getModifiers()&java.lang.reflect.Modifier.STATIC)!=0;
   String instanceClass;
   
  if (specialDispatch) {
            mw.visitVarInsn(ALOAD, 0);
            loadAllArguments(mw, g, bodyIndex);
            mw.visitMethodInsn(INVOKESPECIAL, specialDispatchClass.getName().replace('.','/'),
                     mt.getName(), methodSignature(mt));
   } else if(!statik) {
    instanceClass=generateInstanceAccess(mw,className,mt,g.comboClasses,comboMangled);
    loadAllArguments(mw, g, bodyIndex);
    mw.visitMethodInsn(INVOKEVIRTUAL, instanceClass,
      mt.getName(),methodSignature(mt));
   } else {
    instanceClass=findClassAccess(mw,className,mt,g.comboClasses,comboMangled);
    loadAllArguments(mw, g, bodyIndex);
    mw.visitMethodInsn(INVOKESTATIC, instanceClass,    
      mt.getName(),methodSignature(mt));
   }
   genReturn(mw,mt.getReturnType());
  }


  // must skip raw arguments!
  private static void loadAllArguments(CodeVisitor mw, MethodGroup g, int bodyIndex) {
   Class[] params=g.proto.getParameterTypes();
   Class[] args=g.bodies[bodyIndex].getParameterTypes();
   int arity=params.length;
   int slotIndex=1;
   int argNum=0;
   for (int k=0;k<arity;k++) {
    if (g.protoInfo.raw==null || g.protoInfo.raw[k]==null) {
     Class cc=params[k];
     slotIndex=loadArgument(mw, slotIndex, cc);
 
     // note: there might be upcasts here, that CAN BE REMOVED!
     // the specific method invoked is decided by the signature below.
     if (!(cc.isPrimitive()||Utils.isEqualOrSub(cc,args[argNum]))) {
      mw.visitTypeInsn(CHECKCAST, args[argNum].getName().replace('.','/'));
     }
     argNum++;
    } else slotIndex=skipItem(slotIndex, params[k]);
   }
  }

  private static int loadArgument(CodeVisitor mw, int slotindex, Class cc) {
   if (cc.isPrimitive()) {
    if (cc==byte.class)
     mw.visitVarInsn(ILOAD, slotindex++);      
    else if(cc==char.class)
     mw.visitVarInsn(ILOAD, slotindex++);      
    else if(cc==double.class)
     { mw.visitVarInsn(DLOAD, slotindex++); slotindex++; }
    else if(cc==float.class)
     mw.visitVarInsn(FLOAD, slotindex++);      
    else if(cc==int.class)
     mw.visitVarInsn(ILOAD, slotindex++);      
    else if(cc==long.class)
     { mw.visitVarInsn(LLOAD, slotindex++); slotindex++; }
    else if(cc==short.class)
     mw.visitVarInsn(ILOAD, slotindex++);      
    else if(cc==boolean.class)
     mw.visitVarInsn(ILOAD, slotindex++);      
    else {
     throw new InternalError(0x3CC,   
      "Internal error: unknown class in loadArgument()"
     );
    }
   } else {
    mw.visitVarInsn(ALOAD, slotindex++);
   }
   return slotindex;
  }
  
}











/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////

// An example of the dynamically generated classes:
// Not shown: custom invocation policy and remapNull

/*


// Decompiled by Jad v1.5.8f. Copyright 2001 Pavel Kouznetsov.
// Jad home page: http://www.kpdus.com/jad.html
// Decompiler options: packimports(3) fieldsfirst noctor 

package ovm.polyd;

import java.util.Iterator;
import java.util.List;
import ovm.polyd.policy.Dispatching;
import ovm.polyd.test14.ComboA;
import ovm.polyd.test14.ComboB;
import ovm.polyd.test14.ComboC;
import ovm.polyd.test14.DanceAux1;
import ovm.polyd.test14.Disco;
import ovm.polyd.test14.Factory;
import ovm.polyd.test14.Park;
import ovm.polyd.test14.Person;
import ovm.polyd.test14.Place;
import ovm.polyd.test14.Worker;
import ovm.polyd.test14.Workplace;

// Referenced classes of package ovm.polyd:
//            MultiHash2, MethodGroup, MultiHash1

public final class PolyD$A0
    implements DanceAux1
{

    ComboA instance_ovm_polyd_test14_ComboA_1;
    ComboB instance_ovm_polyd_test14_ComboB_2;
    ComboC instance_ovm_polyd_test14_ComboC_3;
    private static final int quick_map_0[];
    private static final Class keys0_map_0[];
    private static final Class keys1_map_0[];
    private static final MultiHash2 map_0;
    private static final int quick_map_1[];
    private static final Class keys0_map_1[];
    private static final MultiHash1 map_1;

    public final void dance(Person person, Place place)
    {
        Class class1;
        Class class2;
        int i = (class1 = ovm/polyd/test14/Person).hashCode() * 37 + (class2 = place.getClass()).hashCode() & 0x7f;
        int j = class1 != keys0_map_0[i] || class2 != keys1_map_0[i] ? map_0.findCode(i, class1, class2) : quick_map_0[i];
        switch(j != 0 ? j : map_0.lookupAndUpdate(i, class1, class2))
        {
        case 0: // '\0'
        default:
            map_0.methGroup.onMissing(new Class[] {
                class1, class2
            });
            return;

        case 1: // '\001'
            instance_ovm_polyd_test14_ComboA_1.dance(person, (Park)place);
            return;

        case 2: // '\002'
            instance_ovm_polyd_test14_ComboA_1.dance(person, (Workplace)place);
            return;

        case 3: // '\003'
            instance_ovm_polyd_test14_ComboA_1.dance((Worker)person, (Park)place);
            return;

        case 4: // '\004'
            instance_ovm_polyd_test14_ComboA_1.dance(person, place);
            return;

        case 5: // '\005'
            instance_ovm_polyd_test14_ComboB_2.dance(person, (Disco)place);
            return;

        case 6: // '\006'
            instance_ovm_polyd_test14_ComboB_2.dance((Worker)person, (Disco)place);
            return;

        case 7: // '\007'
            instance_ovm_polyd_test14_ComboB_2.dance((Worker)person, (Workplace)place);
            break;
        }
    }

    public final void visit(Place place)
    {
        Class class1;
        int i = (class1 = ovm/polyd/test14/Place).hashCode() & 0x7f;
        int j = class1 != keys0_map_1[i] ? map_1.findCode(i, class1) : quick_map_1[i];
        switch(j != 0 ? j : map_1.lookupAndUpdate(i, class1))
        {
        case 0: // '\0'
        default:
            map_1.methGroup.theDispatcher.onMissing(new Class[] {
                class1
            }, map_1.methGroup.bodies);
            return;

        case 1: // '\001'
            instance_ovm_polyd_test14_ComboA_1.visit(place);
            return;

        case 2: // '\002'
            instance_ovm_polyd_test14_ComboA_1.visit((Factory)place);
            return;

        case 3: // '\003'
            ComboB.visit((Park)place);
            return;

        case 4: // '\004'
            ComboB.visit((Disco)place);
            break;
        }
    }

    public static final void SPECIAL$$STATIC$$INIT(List list)
    {
        Iterator iterator = list.iterator();
        map_0 = new MultiHash2(quick_map_0 = new int[128], keys0_map_0 = new Class[128], keys1_map_0 = new Class[128], (MethodGroup)iterator.next());
        map_1 = new MultiHash1(quick_map_1 = new int[128], keys0_map_1 = new Class[128], (MethodGroup)iterator.next());
    }

}





// Decompiled by Jad v1.5.8f. Copyright 2001 Pavel Kouznetsov.
// Jad home page: http://www.kpdus.com/jad.html
// Decompiler options: packimports(3) fieldsfirst noctor 

package ovm.polyd;

import ovm.polyd.test14.ComboA;
import ovm.polyd.test14.ComboB;
import ovm.polyd.test14.ComboC;

// Referenced classes of package ovm.polyd:
//            Factory

public final class Factory$A0 extends Factory
{

    public Object getDispatcher3(Object obj, Object obj1, Object obj2)
    {
        PolyD.A0 a0 = new PolyD.A0();
        a0.instance_ovm_polyd_test14_ComboA_1 = (ComboA)obj;
        a0.instance_ovm_polyd_test14_ComboB_2 = (ComboB)obj1;
        a0.instance_ovm_polyd_test14_ComboC_3 = (ComboC)obj2;
        ((ComboA)obj).super1 = a0;
        ((ComboC)obj2).aux1 = a0;
        return a0;
    }

}

*/
