/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd;
import ovm.polyd.Missing;

import java.util.Map;
import java.util.List;
import java.lang.reflect.*;

import ovm.polyd.policy.Dispatching;
import ovm.polyd.policy.Invocation;
import ovm.polyd.PolyD.DispatcherSignature;

/**
 * Accumulates information about the new dispatcher. Use a <code>Descriptor</code> to build dispatchers manually,
 * rather than using annotations.
 * <p>
 * This class is not normally accessed directly, and it is of use
 * only when using the pre-5.0
 * version of PolyD. Here is a quick example of its use, please refer to
 * the manual for more information.<p>
 * <code>Descriptor d1=new Descriptor(Interf.class,new Class[] {BodyA.class,BodyB.class});<br>
 * d1.setDispatching(MultiDisp.class);<br>
 * mt=Dance.class.getMethod("visit",new Class[]{Place.class});<br>
 * d1.setMissingHandling(mt,Missing.Ignore);<br>
 * d1.setMethodNullDefaults(mt3,new Class[]{null,Object.class,null});<br>
 * f=Body.class.getField("self");<br>
 * d1.setSelfField(f);<br>
 * <br>
 * Factory fact=d1.register();<br>
 * Interf disp1=fact.getDispatcher1(bod1);<br>
 * <br>
 * disp1.theMessage(arg1,arg2);</code><p>
 * In order to create a dispatcher manually, first create a new descriptor
 * specifying the interface and the list of classes that will be used by
 * the bodies. Then set the various attributes for the whole interface.
 * You can then add individual attributes for the messages, and prepare the
 * "self" fields.
 * <p>
 * Once the descriptor is ready, register it obtaining a factory. Then
 * instantiate the factory as many times as needed, obtaining the required
 * dispatchers.
 * 
 * @author Antonio Cunei
 * @version $Revision: 1.15 $, $Date: 2005/04/03 02:49:23 $
 * @see Factory
 */
public class Descriptor<T> {

 DispatcherSignature<T> theSignature;
 Dispatching theDispatching;
 Missing theMissing;
 Invocation theInvocation;

 Map<Method,MethInfo> methInfo; // accumulates method-specific info
 List<Field> selfList; // registered self fields

 
 // All arrays in MethInfo are the same size as
 // the list of parameters in the prototype, rather than
 // the list of parameters in the bodies. In other
 // words, raw arguments are included.
 static class MethInfo { // package visibility
  public Dispatching theDispatching;
  Missing theMissing;
  public Invocation theInvocation;

  String theName;
  // Elements are non-null if @As is used on that argument
  Class[] asClasses;
  // Elements are non-null if @IfNull is used on that argument
  Class[] nullDefaults;
  // Elements are non-null if @Preload is used on that argument
  Object[][] preload;
  // Elements are non-null if @Raw is used on that argument
  Class[] raw;
  // Each of the arrays above can be null.

  // The following is initialized lazily, acting as a cache.
  // It stores the list of classes
  // of the method parameters, excluding the raw ones.
  private Class[] filteredParams;

  Class[] getFiltered(Class[] pars) {
   if (raw!=null) {
    // I keep a local cache
    if (filteredParams!=null)
     return filteredParams;
    // must exclude raw params
    int valid=0;
    for (int i=0;i<raw.length;i++)
     if (raw[i]==null) valid++;
    Class[] goodParams=new Class[valid];
    valid=0;
    for (int i=0;i<raw.length;i++)
     if (raw[i]==null) goodParams[valid++]=pars[i];
    filteredParams=goodParams;
    return goodParams;
   } else return pars;
  }

  // as above, but do not cache,
  // and include classes forced by @As
  Class[] getForcedFiltered(Class[] pars) {
   Class[] forcedPars=pars.clone();
   for (int ix=0;ix<pars.length;ix++)
    if (asClasses!=null && asClasses[ix]!=null)
     forcedPars[ix]=asClasses[ix];
   // must filter
   if (raw!=null) {
    // must exclude raw params
    int valid=0;
    for (int i=0;i<raw.length;i++)
     if (raw[i]==null) valid++;
    Class[] goodParams=new Class[valid];
    valid=0;
    for (int i=0;i<raw.length;i++)
     if (raw[i]==null) goodParams[valid++]=forcedPars[i];
    forcedPars=goodParams;
   }
   return forcedPars;
  }

 }
 
 //-----------------------------

 // public version (DispatcherSignature is not exposed)
 /**
  * Creates a new <code>Descriptor</code>, given the interface and
  * the list of classes that will be used by the bodies.
  * 
  * @param interf  the PolyD interface
  * @param bodies  the classes that will be used by bodies when building
  */
 public Descriptor (Class<T> interf,Class[] bodies) {
  this(new DispatcherSignature<T>(interf,bodies));
 }
 
 // package only, does the same as the above method,
 // except it uses a DispatcherSignature, built by PolyD
 Descriptor (DispatcherSignature<T> ds) {
  // verify that the interface is an interface for real.
  if (ds.theInterface==null) {
   throw new DispatcherCreationException(0x601,
    "First argument to build() is null."
   );
  }
  if (!ds.theInterface.isInterface()) {
   throw new DispatcherCreationException(0x602,
    "First argument to build() is not an interface."
   );
  }
  if (ds.theComboClasses.length==0) {
   throw new DispatcherCreationException(0x603,
    "No bodies were specified while building the constructor."
   );
  }
  theSignature=ds;
  methInfo=new java.util.HashMap<Method,MethInfo>();
  selfList=new java.util.LinkedList<Field>();
  theMissing=Missing.STANDARD;
 }
 
 /**
  * Sets the default dispatching policy of this <code>Descriptor</code>
  * 
  * @param policyClass  the policy, must be a subclass of {@link ovm.polyd.policy.Dispatching}
  */
 public void setDispatching(Class policyClass) {
  if (!ovm.polyd.policy.Dispatching.class.isAssignableFrom(policyClass))
   throw new DispatcherCreationException(0x624,
    "The policy "+policyClass.toString()+" must be a subclass of Dispatching"
   );
  try {
   Method m=policyClass.getMethod("theDispatcher");
   theDispatching=(Dispatching)(m.invoke(null));
  } catch(Exception e) {
   throw new DispatcherCreationException(0x604,
    "The dispatching policy "+policyClass.toString()+" could not\n"+
    "  be used for "+theSignature.theInterface.toString()+": unknown policy?"
   );
  }
 }

 /**
  * Sets the default invocation policy of this <code>Descriptor</code>
  * 
  * @param invocationClass  the policy, must be a subclass of {@link ovm.polyd.policy.Invocation}
  */
 public void setInvocation(Class invocationClass) {
  if (!ovm.polyd.policy.Invocation.class.isAssignableFrom(invocationClass))
   throw new DispatcherCreationException(0x625,
     "The policy "+invocationClass.toString()+" must be a subclass of Invocation"
   );
  try {
   Method m=invocationClass.getMethod("theInvocation");
   theInvocation=(Invocation)(m.invoke(null));
  } catch(Exception e) {
   throw new DispatcherCreationException(0x605,
    "The invocation policy "+invocationClass.toString()+" could not\n"+
    "  be used for "+theSignature.theInterface.toString()+": unknown policy?"
   );
  }
 }

 /**
  * Sets the default missing method handling of this <code>Descriptor</code>
  *  
  * @param mi the way in which missing methods should be dealt with
  */
 public void setMissingHandling(Missing mi) {
  theMissing=mi;
 }

 //-----------------------------
 // More checks should be made in the methods below,
 // Specifically, we should check that the methods
 // and fields specified are part of the signature used
 // to build the descriptor.


 /**
  * Sets the dispatching policy of a single prototype.
  * 
  * @param mt the desired prototype in the interface
  * @param policyClass  the policy, must be a subclass of {@link ovm.polyd.policy.Dispatching}
  */
 public void setMethodDispatching(Method mt,Class policyClass) {
  if (!ovm.polyd.policy.Dispatching.class.isAssignableFrom(policyClass))
   throw new DispatcherCreationException(0x626,
    "The policy "+policyClass.toString()+" must be a subclass of Dispatching"
   );
  MethInfo mi=methInfo.get(mt);
  if (mi==null) {
   mi=new MethInfo();
   methInfo.put(mt,mi);
  }
  try {
   Method m=policyClass.getMethod("theDispatcher");
   mi.theDispatching=(Dispatching)(m.invoke(null));
  } catch(Exception e) {
   throw new DispatcherCreationException(0x606,
    "The dispatching policy "+policyClass.toString()+" could not\n"+
    "  be used for prototype "+mt.toString()+": unknown policy."
   );
  }
 }

 /**
  * Sets the invocation policy of a single prototype.
  * 
  * @param mt the desired prototype in the interface
  * @param invocationClass  the policy, must be a subclass of {@link ovm.polyd.policy.Invocation}
  */
 public void setMethodInvocation(Method mt,Class invocationClass) {
  if (!ovm.polyd.policy.Invocation.class.isAssignableFrom(invocationClass))
   throw new DispatcherCreationException(0x627,
     "The policy "+invocationClass.toString()+" must be a subclass of Invocation"
   );
  MethInfo mi=methInfo.get(mt);
  if (mi==null) {
   mi=new MethInfo();
   methInfo.put(mt,mi);
  }
  try {
   Method m=invocationClass.getMethod("theInvocation");
   mi.theInvocation=(Invocation)(m.invoke(null));
  } catch(Exception e) {
   throw new DispatcherCreationException(0x607,
    "The invocation policy "+invocationClass.toString()+" could not\n"+
    "  be used for prototype "+mt.toString()+": unknown policy."
   );
  }
 }

 /**
  * Sets the missing method handling of a single prototype.
  *  
  * @param mt the desired prototype in the interface
  * @param mh the way in which missing methods should be dealt with
  */
  public void setMethodMissingHandling(Method mt,Missing mh) {
  MethInfo mi=methInfo.get(mt);
  if (mi==null) {
   mi=new MethInfo();
   methInfo.put(mt,mi);
  }
  mi.theMissing=mh;
 }

 /**
  * Sets the alternate name of a prototype.
  *  
  * @param mt the desired prototype in the interface
  * @param nm the new name
  */
  public void setMethodName(Method mt,String nm) {
  MethInfo mi=methInfo.get(mt);
  if (mi==null) {
   mi=new MethInfo();
   methInfo.put(mt,mi);
  }
  mi.theName=nm;
 }

 /**
  * Sets the list of forced classes of a prototype.<p>
  * The <code>as</code> argument to this routine is an
  * array of classes, of size equal to the number of
  * parameters of the prototype. The non-null elements
  * of the array will alter the interpretation of the
  * dynamic class of the corresponding prototype argument
  * during dispatching. For each non-null element, the
  * dynamic class of the argument will be ignored, and
  * the element of the array will be used instead in
  * order to select the most appropriate method.
  *  
  * @param mt the desired prototype in the interface
  * @param as the list of forced classes
  */
 public void setMethodAsClasses(Method mt,Class[] as) {
  MethInfo mi=methInfo.get(mt);
  if (mi==null) {
   mi=new MethInfo();
   methInfo.put(mt,mi);
  }
  Class[] pars=mt.getParameterTypes();
  for (int ix=0;ix<pars.length;ix++) {
   if (as[ix]!=null) {
    if (mi.raw!=null && mi.raw[ix]!=null) {
     throw new DispatcherCreationException(0x608,
      "The @As clause cannot be used with a raw argument,\n"+
      "in "+mt.toString()+" ("+theSignature.theInterface.getName()+")"
     );
    }
    if (!Utils.isEqualOrSub(pars[ix],as[ix])) {
     throw new DispatcherCreationException(0x609,
      "@As clause must specify a class which is equal or superclass\n"+
      "of a parameter, in "+mt.toString()+" ("+theSignature.theInterface.getName()+")"
     );
    }
   }
  }
  mi.asClasses=as;
 }

 
 
 /**
  * Sets the list of raw arguments in the prototype.<p>
  * Arguments that are marked as raw are treated in a special
  * way during dispatching: they will be passed as
  * they are to the selection mechanism, and they will
  * not be used as actual arguments during the call to
  * the selected method.<p>
  * Raw arguments can be used to annotate individual
  * call sites, or to discriminate on more information
  * than just the dynamic class of the message arguments.<p>
  * If an argument is marked as raw, it cannot be
  * forced using the {@link ovm.polyd.tags.As} tag,
  * and it cannot be used together with the
  * {@link ovm.polyd.tags.IfNull} tag. A raw argument can
  * be any object or an <code>int</code>, but no other
  * primitive type.<p>
  * This method must be called before setting As, IfNull, or
  * Preload properties.
  *  
  * @param mt the desired prototype in the interface
  * @param raw an indication of whether each argument is raw or not
  */
 public void setMethodRawClasses(Method mt,boolean[] raw) {
  MethInfo mi=methInfo.get(mt);
  if (mi==null) {
   mi=new MethInfo();
   methInfo.put(mt,mi);
  }
  if (mi.nullDefaults!=null) {
    throw new DispatcherCreationException(0x610,
     "The @Raw tag was detected before the @IfNull tag.\n"+
     "If you're using the pre-5.0 API, you need to\n"+
     "use setMethodRawClasses() before setMethodNullDefaults(),\n"+
     "(otherwise you just encountered\n"+
     " an unexpected internal error).\n"+
     "Error while checking: "+Utils.methodString(mt)
    );
  }
  if (mi.asClasses!=null) {
    throw new DispatcherCreationException(0x611,
     "The @Raw tag was detected before the @As tag.\n"+
     "If you're using the pre-5.0 API, you need to\n"+
     "use setMethodRawClasses() before setMethodAsClasses()\n"+
     "(otherwise you just encountered\n"+
     " an unexpected internal error).\n"+
     "Error while checking: "+Utils.methodString(mt)
    );
  }
  if (mi.preload!=null) {
   throw new DispatcherCreationException(0x661,
    "The @Raw tag was detected before the @Preload tag.\n"+
    "If you're using the pre-5.0 API, you need to\n"+
    "use setMethodRawClasses() before setMethodPreload()\n"+
    "(otherwise you just encountered\n"+
    " an unexpected internal error).\n"+
    "Error while checking: "+Utils.methodString(mt)
   );
 }
  Class[] pars=mt.getParameterTypes();
  Class[] rawC=new Class[pars.length];
  for (int ix=0;ix<pars.length;ix++) {
   if (raw[ix]) {
    if (pars[ix].isPrimitive() && pars[ix]!=int.class) {
     throw new DispatcherCreationException(0x612,
      "Tag @Raw can only be used with Objects or ints!\n"+
      "Error while checking: "+Utils.methodString(mt)
     );
    }
    rawC[ix]=pars[ix];
   }
  }
  mi.raw=rawC;
 }


 /**
  * Sets the list of defaults for null arguments in the prototype.<p>
  * The <code>nul</code> argument to this routine is an
  * array of classes, of size equal to the number of
  * parameters of the prototype. The non-null elements
  * of the array will supply a default interpretation for
  * those arguments that are found, at runtime, to be <code>null</code>.
  *  
  * @param mt the desired prototype in the interface
  * @param nul the list of defaults for null arguments
  */
 public void setMethodNullDefaults(Method mt,Class[] nul) {
  MethInfo mi=methInfo.get(mt);
  if (mi==null) {
   mi=new MethInfo();
   methInfo.put(mt,mi);
  }
  Class[] pars=mt.getParameterTypes();
  for (int ix=0;ix<pars.length;ix++) {
   if (nul[ix]!=null)
   if (mi.raw!=null && mi.raw[ix]!=null) {
    throw new DispatcherCreationException(0x613,
     "Cannot use @IfNull with a raw argument\n"+
     "in "+mt.toString()+" ("+theSignature.theInterface.getName()+")"
    );
   } else { // not raw
    if (!Utils.isEqualOrSub(nul[ix],pars[ix])) {
     throw new DispatcherCreationException(0x614,
      "@IfNull clause must specify a class which is equal or subclass\n"+
      "of a parameter, in "+mt.toString()+" ("+theSignature.theInterface.getName()+")"
     );
    }
   }
  }
  mi.nullDefaults=nul;
 }

 /**
  * Sets the list of defaults for null arguments in the prototype.<p>
  * The <code>nul</code> argument to this routine is an
  * array of classes, of size equal to the number of
  * parameters of the prototype. The non-null elements
  * of the array will supply a default interpretation for
  * those arguments that are found, at runtime, to be <code>null</code>.
  * Preload can be used for raw arguments as well, but due
  * to restrictions in the handling of annotations by Java,
  * only raw arguments that happen to be classes can be
  * initialized using annotations. Using this routine, no
  * limitations exist.
  * @param mt the desired prototype in the interface
  * @param prel the array of arrays of arguments that should be preloaded
  */
 public void setMethodPreload(Method mt,Object[][] prel) {
  MethInfo mi=methInfo.get(mt);
  if (mi==null) {
   mi=new MethInfo();
   methInfo.put(mt,mi);
  }
  Class[] pars=mt.getParameterTypes();
  for (int iy=0;iy<prel.length;iy++) {
   // if raw, each element must be of class equal or sub of that
   // of the raw parameter.
   // if not raw, consider the element rather than its class
   Object[] objectList=prel[iy];
   if (objectList.length!=pars.length) {
    throw new DispatcherCreationException(0x615,
     "Mismatched length in @Seq annotation: must have "+pars.length+" elements\n"+
     "in "+mt.toString()+" ("+theSignature.theInterface.getName()+")"
    );
   }
   for (int ix=0;ix<objectList.length;ix++) {
    if (mi.raw==null || mi.raw[ix]==null) {
     if (!(objectList[ix] instanceof Class) || !Utils.isEqualOrSub((Class)(objectList[ix]),pars[ix])) {
      throw new DispatcherCreationException(0x616,
       "@Preload clause must specify classes which are equal or subclasses\n"+
       "of parameters, in "+mt.toString()+" ("+theSignature.theInterface.getName()+")"
      );
     }
    } else { // raw
     Class elCl=objectList[ix].getClass();
     if (elCl.isPrimitive() && elCl!=int.class) {
      throw new DispatcherCreationException(0x676,
        "Cannot preload a raw argument with a non-int primitive"+
        " in "+mt.toString()+" ("+theSignature.theInterface.getName()+")"
       );
     }
     if (!Utils.isEqualOrSub(elCl,pars[ix])) {
      throw new DispatcherCreationException(0x677,
        "Cannot preload a raw argument: incompatible object"+
        " in "+mt.toString()+" ("+theSignature.theInterface.getName()+")"
       );
     }
    }
   }
  }
  mi.preload=prel;
 }

 /**
  * Marks this field as a self field.<p>
  * The field will be automatically initialized at dispatcher-creation time.
  * @param f The self field, contained in one of the bodies
  */
 public void setSelfField(Field f) {
  selfList.add(f);
 }
 
 
 /**
  * Generates a new <code>Factory</code>, registering this <code>Descriptor<code>.<p>
  * The factory can then be used to create new dispatchers.
  * @return the new <code>Factory</code>
  */
 public Factory register() {
  return PolyD.register(this,false);
 }
 
 /**
  * Registers this <code>Descriptor</code> as a special dispatcher
  * (see manual for details).
  * This method could be removed from future versions of PolyD,
  * and its use is not recommended.
  * @deprecated Use {@link #register} instead.
  */
 public void registerSpecial() {
  PolyD.register(this,true);
 }
 
}


