/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd;

import java.lang.reflect.Method;

import ovm.polyd.policy.Dispatching;
import ovm.polyd.policy.Invocation;
import ovm.polyd.Descriptor;
import ovm.polyd.Descriptor.MethInfo;
/*#ifdef14
#else14 */
import ovm.polyd.tags.Name;
//#endif14
import ovm.polyd.Missing;



final class MethodGroup {
 // includes raw arguments
 int arityRaw; // the same as proto.getparameterTypes().length
 // to get the arity without raw args, use getFiltered().length
 String name; // the same as findName()

 Method proto;
 public MethInfo protoInfo; // useful info on the prototype
 public Method[] bodies;

 // classes used to build the dispatcher:
 Class[] comboClasses;

 DispatchInfo dispatchInfoNeeded; // set IF a stub for this group
 // of prototypes was created while generating the master class,
 // and therefore a HashMap was used. Defaults to NONE.
 //
 // indicates whether the bytecode generated in the master class
 // to represent a method that correspond to this MethodGroup
 // there is the need for some local information (an instance
 // variable) in order to retrive some information necessary
 // for the dispatching, at runtime.
 //
 // in some cases the dispatching info is necessary, but
 // a hash map is not used. use METHODGROUP in that case
 //
/*#ifdef14
    public static final class DispatchInfo
    {

        public static final DispatchInfo NONE;
        public static final DispatchInfo MAP;
        public static final DispatchInfo METHODGROUP;
        private String name;

        public String toString()
        {
            return (name);
        }

        static
        {
            NONE = new DispatchInfo("NONE");
            MAP = new DispatchInfo("MAP");
            METHODGROUP = new DispatchInfo("METHODGROUP");
        }

        private DispatchInfo(String s)
        {
            name=s;
        }
    }
#else14 */
 public enum DispatchInfo {
  NONE, // direct dispatch, no info required
  MAP,  // a map is used, methodgroup can be accessed as a field
  METHODGROUP // the local var stores directly the methodgroup
 }
//#endif14
 
 public String toString() {
  String s;
  s="\n---\nMethodGroup:\n";
  s+="Name: "+name+"\n";
  if (arityRaw==getFiltered().length) {
   s+="Arity: "+arityRaw+"\n";
  } else {
   s+="Arity: "+getFiltered().length+"\n";
   s+="Arity (with raw arguments): "+arityRaw+"\n";
  }
  s+="Proto:\n";
  s+=Utils.methodString(proto)+"\n";
  if (bodies!=null) {
   s+="Bodies:\n";
   for (Method m : bodies)
    s+=Utils.methodString(m)+"\n";
  } else s+="No bodies\n";
  s+="---";
  return s;
 }
 


// refines getFiltered in MethodGroup
 Class[] getFiltered() {
  return protoInfo.getFiltered(proto.getParameterTypes());
 }

// as above, but include classes forced by @As
 Class[] getForcedFiltered() {
  return protoInfo.getForcedFiltered(proto.getParameterTypes());
 }

 public <T> MethodGroup (Method proto,Method[] bodies,Descriptor<T> dd) {
  dispatchInfoNeeded=DispatchInfo.NONE;
  this.bodies=bodies;
  this.proto=proto;
  this.name=PolyD.findName(dd,proto);
  this.arityRaw=proto.getParameterTypes().length;
  MethInfo mi=dd.methInfo.get(proto);
  if (mi==null) mi=new MethInfo();
  if (mi.theDispatching==null)
       mi.theDispatching=dd.theDispatching; // might be null
  if (mi.theInvocation==null)
       mi.theInvocation =dd.theInvocation;  // might be null
  if (mi.theMissing==null)
       mi.theMissing    =dd.theMissing;     // might be null
  this.protoInfo=mi;
  this.comboClasses=dd.theSignature.theComboClasses;
 }

//
// the methods below are glue code between the MultiHashx and the dispatching policy

 // warning: whereas bestMatch returns -1,0..n-1,
 // this routine returns 0,1..n
 int findBestMatch(Object[] oargs) {
  Split split=new Split(oargs);
  return protoInfo.theDispatching.bestMatch(split.args,bodies,split.raws,comboClasses,proto,name)
         +1; // !!!
 }

 public void onMissing(Object[] oargs) {
  Split split=new Split(oargs);
  if (!protoInfo.theDispatching.handleMissing(split.args,bodies,split.raws,comboClasses,proto,name)) {
   String s="The following message:\n";
   s+=proto.getName()+"(";
   String sep="";
   int ra=0;
   int ar=0;
   for (int i=0;i<oargs.length;i++) {
    if (protoInfo.raw==null || protoInfo.raw[i]==null) {
     s+=sep+split.args[ar++].getName();
    } else {
     s+=sep+"@Raw "+split.raws[ra++];
     sep=",";
    }
   }
   s+=")";
   if (!proto.getName().equals(name))
    s+="  (renamed as "+name+")";
   s+="\ncould not be dispatched according to policy: "+
      protoInfo.theDispatching.toString();
   throw new MissingMethodException(s);
  }
 }

 private class Split {
  Class[] args;
  Object[] raws;

  Split(Class[] args,Object[] raws) {
   this.raws=raws;
   this.args=args;
  }
  Split(Object[] oargs) {
  Class[] raw=protoInfo.raw;
  // split oargs in args and raw
  int ra=0;
  int ar=0;
  for (int i=0;i<oargs.length;i++)
   if (raw==null || raw[i]==null) ar++; else ra++;
  raws=new Object[ra];
  args=new Class[ar];
  ar=0; ra=0;
  for (int i=0;i<oargs.length;i++) {
   if (raw==null || raw[i]==null) {
    if (!(oargs[i] instanceof Class)) {
     throw new InternalError(0x5C1,
      "Internal error while splitting arguments."
     );
    }
    args[ar++]=(Class)(oargs[i]);    
   } else {
    raws[ra++]=oargs[i];
   }
  }
 }

 }

 public Object[] remapNull(Object[] oargs) {
  Split split=new Split(oargs);
  Object[][] remapped=protoInfo.theDispatching.remapNull(split.args,bodies,split.raws,comboClasses,proto,name);
  Object[] args=remapped[0];
  Object[] raws=remapped[1];
  // and merge them again
  int ar=0;
  int ra=0;
  for (int i=0;i<oargs.length;i++) {
   if (protoInfo.raw==null || protoInfo.raw[i]==null) {
    oargs[i]=args[ar++];
   } else {
    oargs[i]=raws[ra++];
   }
  }
  return oargs;
 }

 public Method[] compatibleSet() {
  return protoInfo.theDispatching.compatibleSet(getFiltered(),bodies,comboClasses,proto,name);
 }


}
