/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd;

import ovm.polyd.Missing;

/*
 * This map is initialized lazily. Each class combination
 * is cached, so that over time the map ends up being only
 * read, and very rarely updated.
 * 
 * Updates are performed in a synchronized section, but the
 * reads are not synchronized. That means that part of the
 * data structures might not appear consistent to all threads.
 * However, the worst that can happen to threads is that
 * they do not find an object combination that should be
 * there, or they see a "value" of zero while a non-zero
 * value is being written.
 * 
 * In every case, the worst that can happen is that the
 * value "zero" is found (part of the logic of this map
 * actually ends up being embedded in the stub in the
 * master class. Check the dynamically generated code
 * for more details). If zero is returned, it might be
 * that a method was not found, or that a different
 * thread is updating (in a synchronized section), so
 * that the proper value has not yet been written, or
 * the keys have not been completely written.
 * 
 * The handler for the case "zero" in the switch (in the
 * stub) will have to check again, calling a synchronized
 * "lookup-and-update" that performs the lookup again,
 * in a synchronized fashion this time, and updates
 * the maps if necessary.
 * 
 * During normal operation, such a mechanism offers good
 * performance with no synchronization (just reads). There
 * is some overhead when a new class (object) combination is
 * encountered, but the combination is cached from that
 * moment on, so it's not too bad. There is also a double
 * lookup when a method really is not found. Actually,
 * the "method not found" condition is *not* cached at all,
 * so that case is rather expensive. We will assume, though,
 * that it is an uncommon error condition, so performance
 * should actually not be a concern in that particular case.
 */

/**
 * @author Antonio Cunei
 * @version $Revision: 1.12 $, $Date: 2005/04/03 00:08:59 $
 */
final class MultiHash1  {
 
 private int[] values;           // array of values
 private Node[] collisions;
 public MethodGroup methGroup;
 private Object[] keys_1;

 public MultiHash1(int[] values_,Object[] keys1_,MethodGroup methGroup_) {
  values = values_;
  methGroup=methGroup_;
  collisions = new Node[values.length];
  keys_1=keys1_;
  if (methGroup_.protoInfo.preload!=null) {
   preload(methGroup_.protoInfo.preload);
   methGroup_.protoInfo.preload=null; // GC
  }
 }
 
 static private final class Node {
  Node link;
  int value;
  Object key_1;
  Node(Node l, Object key1, int v) {
   link = l;
   value = v;
   key_1 = key1;
  }
 }
 
 
 int findBody(Object key1) {
  
  // returns index in bodies
  
  Object[] args=new Object[] {key1};
  return methGroup.findBestMatch(args);
 }

 
 public int findCode(int hash,Object key1) {
  // If here, must look in the collision chain
  Node p, h = collisions[hash];
  for (p = h; p != null; p = p.link) {
   if (key1==p.key_1) {
    return p.value;
   }
  }
  return 0;
 }

 
 public synchronized int lookupAndUpdate(int hash,Object key1) {

  if (keys_1[hash]==key1) return values[hash];

  // If here, either collision or values[hash] was zero.
  
  if (keys_1[hash] == null) { // simple insert

   keys_1[hash] = key1;
   
   // at this point this combination could have a body definition or not.
   // let's look for the best one
   
   values[hash] = findBody(key1);
   return values[hash];
  }
  
  Node p, h = collisions[hash];
  for (p = h; p != null; p = p.link) {
   if (key1==p.key_1) {
    return p.value;
   }
  }
  
  int value=findBody(key1);
  collisions[hash] = new Node(h, key1, value); 
  return value;
 }
 
 // the hashing algorithm used here MUST ABSOLUTELY MUST be identical
 // to the one included in the automatically generated bytecode
 public void preload(Object[][] pre) {
  int mask=(PolyD.guessTableSize(methGroup)-1);
  Missing missing=methGroup.protoInfo.theMissing;
  for (Object[] v : pre) {
   int hash=v[0].hashCode()&mask;
   if (lookupAndUpdate(hash,v[0])==0) {
    if (missing==Missing.ABORT) {
     System.err.println("Fatal Error - Selector didn't find a suitable method.\nAborting.");
     System.exit(10);
    } else
    if (missing==Missing.WARN) {
     System.err.println("Warning - Selector didn't find a suitable method.");
    } else
    if (missing==Missing.FAIL) {
     throw new MissingMethodException();
    } else
    if (missing==Missing.STANDARD) {
     methGroup.onMissing(v);
    } else
    if (missing==Missing.IGNORE) {
    } else {
     throw new InternalError(0x201,
      "Internal error - unknown Missing Method handling."
     );
    }
   }
  }
 }
 
}




