/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd;

import ovm.polyd.Missing;

/**
 * @author Antonio Cunei
 * @version $Revision: 1.12 $, $Date: 2005/04/03 00:08:59 $
 */
final class MultiHash4 {
 
 private int[] values;           // array of values
 private Node[] collisions;
 public MethodGroup methGroup;
 private Object[] keys_1;
 private Object[] keys_2;
 private Object[] keys_3;
 private Object[] keys_4;

 public MultiHash4(int[] values_,Object[] keys1_,Object[] keys2_,
   Object[] keys3_,Object[] keys4_,MethodGroup methGroup_) {
  values = values_;
  methGroup=methGroup_;
  collisions = new Node[values.length];
  keys_1=keys1_;
  keys_2=keys2_;
  keys_3=keys3_;
  keys_4=keys4_;
  if (methGroup_.protoInfo.preload!=null) {
   preload(methGroup_.protoInfo.preload);
   methGroup_.protoInfo.preload=null; // GC
  }
 }
 
 static private final class Node {
  Node link;
  int value;
  Object key_1;
  Object key_2;
  Object key_3;
  Object key_4;
  Node(Node l, Object key1, Object key2, Object key3, Object key4, int v) {
   link = l;
   value = v;
   key_1 = key1;
   key_2 = key2;
   key_3 = key3;
   key_4 = key4;
  }
 }
 
 int findBody(Object key1,Object key2,Object key3,Object key4) {
  
  // returns index in bodies
  
  Object[] args=new Object[] {key1,key2,key3,key4};
  return methGroup.findBestMatch(args);
 }
 
 public int findCode(int hash,Object key1,Object key2,Object key3,Object key4) {
  // If here, must look in the collision chain
  Node p, h = collisions[hash];
  for (p = h; p != null; p = p.link) {
   if (key1==p.key_1&&key2==p.key_2&&key3==p.key_3&&key4==p.key_4) {
    return p.value;
   }
  }
  return 0;
 }

 public synchronized int lookupAndUpdate(int hash,Object key1,Object key2,Object key3,Object key4) {

  if (keys_1[hash]==key1&&keys_2[hash]==key2&&
      keys_3[hash]==key3&&keys_4[hash]==key4) return values[hash];

  // If here, either collision or values[hash] was zero.
  
  if (keys_1[hash] == null) { // simple insert

   keys_1[hash] = key1;
   keys_2[hash] = key2;
   keys_3[hash] = key3;
   keys_4[hash] = key4;
   
   // at this point this combination could have a body definition or not.
   // let's look for the best one
   
   values[hash] = findBody(key1,key2,key3,key4);
   return values[hash];
  }
  
  Node p, h = collisions[hash];
  for (p = h; p != null; p = p.link) {
   if (key1==p.key_1&&key2==p.key_2&&key3==p.key_3&&key4==p.key_4) {
    return p.value;
   }
  }
  
  int value=findBody(key1,key2,key3,key4);
  collisions[hash] = new Node(h, key1, key2, key3, key4, value); 
  return value;
 }
 
 // the hashing algorithm used here MUST ABSOLUTELY MUST be identical
 // to the one included in the automatically generated bytecode
 // Similarly, handling of MissingMethod must reflect what is done
 // in the class generated by ClassBuilder
 
 public void preload(Object[][] pre) {
  int mask=(PolyD.guessTableSize(methGroup)-1);
  Missing missing=methGroup.protoInfo.theMissing;
  for (Object[] v : pre) {
   int hash=(((v[0].hashCode()*37+
                v[1].hashCode())*37+
                 v[2].hashCode())*37+
                  v[3].hashCode())&mask;
   if (lookupAndUpdate(hash,v[0],v[1],v[2],v[3])==0) {
    if (missing==Missing.ABORT) {
     System.err.println("Fatal Error - Selector didn't find a suitable method.\nAborting.");
     System.exit(10);
    } else
    if (missing==Missing.WARN) {
     System.err.println("Warning - Selector didn't find a suitable method.");
    } else
    if (missing==Missing.FAIL) {
     throw new MissingMethodException();
    } else
    if (missing==Missing.STANDARD) {
     methGroup.onMissing(v);
    } else
    if (missing==Missing.IGNORE) {
    } else {
     throw new InternalError(0x204,
      "Internal error - unknown Missing Method handling."  
     );
    }
   }
  }
 }

}

