/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd;

import ovm.polyd.Missing;

/**
 * @author Antonio Cunei
 * @version $Revision: 1.11 $, $Date: 2005/04/03 00:08:59 $
 */
final class MultiHashN {
 
 final private int mask_;            // mask for main arrays
 final private int maskCollisions_;  // mask for collision table
 private Object[][] keys_n;           // array of keys
 private int[] values_;           // array of values
 private Node[] collisions_;
 
 public MethodGroup methGroup;

 public MultiHashN(MethodGroup methGroup_) {
  int sizeCollisions;
  int size;
  int hint=methGroup_.bodies.length*3;
  if (hint<32)
   hint=32; // minimal size.

  
  size = 8;
  while ( size < hint ) size *= 2;
  mask_ = size-1;
  sizeCollisions = size>>3;
  maskCollisions_ = sizeCollisions-1; /* make collision table 1/8th of the size of the main table */
  collisions_ = new Node[sizeCollisions];
  keys_n = new Object[size][];
  values_ = new int[size];
  methGroup=methGroup_;
  if (methGroup_.protoInfo.preload!=null) {
   preload(methGroup_.protoInfo.preload);
   methGroup_.protoInfo.preload=null; // GC
  }
 }
 
 static private final class Node {
  Node link;
  Object[] key_n;
  int value;
  Node(Node l, Object[] keyn, int v) {
   link = l;
   key_n = keyn;
   value = v;
  } 
 }
 
 int findBody(Object[] args) {
  
  // returns index in bodies
  
  return methGroup.findBestMatch(args);
 }
 
 
 public int findCode(Object[] keyn) {
  int hash = 0;
  for (Object c : keyn)
   hash = 37*hash + c.hashCode();
  hash = mask_&hash;
  if (java.util.Arrays.equals(keyn,keys_n[hash])) { // full equals, not ptr identity
   return values_[hash];
  }
  // If here, must look in the collision chain
  hash = maskCollisions_&hash; // find pos in the collision mask
  Node p, h = collisions_[hash];
  for (p = h; p != null; p = p.link) {
   if (java.util.Arrays.equals(keyn,p.key_n)) {
    return p.value;
   }
  }
  return 0;
 }

 public synchronized int lookupAndUpdate(Object[] keyn) {

  int hash = 0;
  for (Object c : keyn)
   hash = 37*hash + c.hashCode();
  hash = mask_&hash;
  if (java.util.Arrays.equals(keyn,keys_n[hash])) {
   return values_[hash];
  }
  
  // if they are different -> either hash is null or collision (keys_1[hash] is null)
  // in the normal case, I save the test hash==null, which I can safely do afterwards
  
  if (keys_n[hash] == null) { // simple insert
   keys_n[hash]=keyn;
   
   // at this point this combination could have a body definition or not.
   // let's look for the best one
   
   values_[hash] = findBody(keyn);
   return values_[hash];
  }
  
  hash = maskCollisions_&hash; // find pos in the collision mask
  Node p, h = collisions_[hash];
  for (p = h; p != null; p = p.link) {
   if (java.util.Arrays.equals(keyn,p.key_n)) {
    return p.value;
   }
  }
  
  
  int value=findBody(keyn);
  collisions_[hash] = new Node(h, keyn, value); 
  return value;
 }
 
 
 // Handling of MissingMethod must reflect what is done
 // in the class generated by ClassBuilder
 
 public void preload(Object[][] pre) {
  int mask=(PolyD.guessTableSize(methGroup)-1);
  Missing missing=methGroup.protoInfo.theMissing;
  for (Object[] v : pre) {
   if (lookupAndUpdate(v)==0) {
    if (missing==Missing.ABORT) {
     System.err.println("Fatal Error - Selector didn't find a suitable method.\nAborting.");
     System.exit(10);
    } else
    if (missing==Missing.WARN) {
     System.err.println("Warning - Selector didn't find a suitable method.");
    } else
    if (missing==Missing.FAIL) {
     throw new MissingMethodException();
    } else
    if (missing==Missing.STANDARD) {
     methGroup.onMissing(v);
    } else
    if (missing==Missing.IGNORE) {
    } else {
     throw new InternalError(0x205,
      "Internal error - unknown Missing Method handling."  
     );
    }
   }
  }
 }

} 

