/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.LinkedList;
import java.util.List;
import java.lang.annotation.Annotation;
import ovm.polyd.Utils.Diff;
import static ovm.polyd.Utils.Diff.*;
import ovm.polyd.policy.Dispatching;
import ovm.polyd.policy.Invocation;
/*#ifdef14
#else14 */
import ovm.polyd.tags.As;
import ovm.polyd.tags.DispatchingPolicy;
import ovm.polyd.tags.IfNull;
import ovm.polyd.tags.InvocationPolicy;
import ovm.polyd.tags.Name;
import ovm.polyd.tags.OnMissing;
import ovm.polyd.tags.Preload;
import ovm.polyd.tags.Self;
import ovm.polyd.tags.Seq;
import ovm.polyd.tags.Raw;
//#endif14

/*
  Welcome to PolyD!

  Assuming you are reading this because you would like to know more
  about the structure and the internals of PolyD, here is some brief
  information that can be useful for you to find your way around here.

  This file is the main module: it contains the crucial build()
  method and its relatives. The actual construction of the dynamically
  generated bytecode, however, is done in the ClassBuilder.

  What PolyD.java contains is the scanning of the annotations, the
  first coarse verification that all prototype have some method
  that matches, and the caching of the generated bytecode.

  The other crucial module, ClassBuilder.java, contains the gritty
  bytecode generation, with all its variations and details.

  Apart from these two, the other modules are fairly short and
  self-explanatory. The MultiHash files contain the implementation
  of a custom kind of hash table, particularly suitable to this
  code since they do not usually require synchronization most of
  the times. Another important module is MethodGroup, which
  stores all of the information related to the expansion of a
  prototype in the interface of a dispatcher: the list of
  compatible bodies, how to deal with raw arguments, and similar
  information.

  The way in which methodgroups and the maps are passed around
  at runtime is a bit peculiar. The dynamically generated code
  stores in its static fields the main arrays that make up the
  hash maps, and a pointer to the hashmap as a whole.

  Initially, The arrays are created and passed to the constructor
  of the hashmap, together with the methodgroup. The hashmap
  keeps locally a reference to the arrays and to the methodgroup.
  A reference to the constructed hashmap is then stored in the
  class as a static field.

  At runtime, the arrays are directly inspected by the synthetic
  class. If no match is found, the reference to the hashmap is
  used to perform the slow-path selection. The hashmap knows
  where are its arrays, and relies on the methodgroup to find
  the rest of the information (in particular the dispatching
  policy). Once the information is found, the arrays are
  updated and from that moment on the hashmap will not be
  used again, unless a collision occours or a new combination
  of argument classes is found.


  Another small note: there are often references in the code
  to "combos". They are the instances of the implementing
  classes, that the manual defines as "bodies". Just to
  further complicate things, in the code there are
  references to "bodies", that are actually the individual
  implementing methods.

  There are also here and there strange single-line comments
  that look like #ifdef14 and #else14. They are used to express
  in-place some alternative version of certain code or data
  necessary in the pre-5.0 version of PolyD.
  During compilation, the stock files are compiled to generate
  1.5 bytecode, and some postprocessing is done in order to
  generate the pre-5.0 version. In particular, a tiny "sed"
  script parses the comments above to replace the code,
  then the 1.5 javac is used with -target jsr14 to generate
  1.4 bytecode, which is then decompiled by the "jad"
  decompiler, and the result is finally recompiled again
  as 1.1 bytecode. Incredibly, everything works perfectly
  after this chain of transformations.

  The pre-5.0 version only uses 1.1 bytecode, and was run
  successfully using Sun's JRE 1.3.1 and also using
  Sun's JRE 1.2.2 (with some minor warnings from the VM).

*/

import ovm.polyd.Missing;

/**
 * The main class of PolyD: use the methods of this class to build new dispatchers.<p>
 * Normally, a new dispatcher is simply built by constructing
 * the desired interface and bodies, using annotations, and calling the
 * static method <code>PolyD.build(...)</code>.<p>
 * A number of auxiliary variations on the main mechanism are also offered,
 * and documented below.
 * For more information, please consult the manual
 * available online at
 * <a href="http://www.ovmj.org/polyd">http://www.ovmj.org/polyd</A>.
 * 
 * @see ovm.polyd.policy.Dispatching
 * @see ovm.polyd.policy.Invocation
 * @version $Revision: 1.21 $, $Date: 2005/04/03 03:02:26 $
 * @author Antonio Cunei
 *
 */
public abstract class PolyD {
 private PolyD() {}
 
 // various implemented features:
 
 // special dispatching:
 //  it allows one body to implement the interface. The result
 //  is that, rather than storing the body instance within the
 //  dispatcher, the dispatcher itself becomes an instance of
 //  a subclass of the class of the body. The advantage is that,
 //  in order to dispatch messages, "super" can be used rather
 //  than redispatching to a different instance. In practical
 //  terms, the increase in speed is negligible, and the feature
 //  could be removed in the future.

 // static methods in bodies:
 //  if a method does not require access to the instance that
 //  implements the body, a static method can be used instead.
 //  It should not make much of a difference: using a static
 //  method is mostly a matter of convenience.

 // preloading:
 //  in order to force an early checking of certain combinations
 //  of classes, and to preload the method caches, preloading
 //  is now supported. The logic is in the MultiHash files.

 // raw arguments:
 //  additional message arguments that are not passed on to
 //  the called method, but are used insted by the method
 //  selector in the dispatching policy

 // exceptions:
 //  all PolyD errors are now handled as exceptions, that can
 //  be conveniently caught by the client code.

 // synchronization:
 //  building new dispatcher is MT-safe. Similarly, the
 //  hash maps used to store the method caches were also
 //  designed to be efficient but safe in a multithreaded
 //  environment. This aspect was not extensively tested,
 //  though, so there might be glitches remaining
 //  (hopefully not)

 // embedded hash access:
 //  in the dynamically generated bytecode, the hashmaps
 //  are directly accessed by the stub code, rather than
 //  by the main hash map class. That allows the code to
 //  be slightly faster, as one virtual invocation per
 //  dispatch is removed (see the sample code at the
 //  end of ClassBuilder for an example)

 // super:
 //  the @As tag implements a more general form of super.
 //  It is also possible to achieve a similar effect by
 //  designing a custom policy.

 // support methods:
 //  bodies can contain methods that are not present
 //  in the interface; they are considered to be
 //  support methods and are conveniently ignored.

 // self:
 //  If a method was called through a certain
 //  dispatcher, this mechanism allows to call
 //  from within that method further methods using
 //  again the same dispatcher (or a different one)

 // covariant return types:
 //  they are allowed, support is implemented

 // (more stuff is described in the manual)

 // TODO:
 //  currently the classes of primitive args are hashed, even
 //  though there would be need to do so: there is no
 //  subsumption. A similar argument applies for final classes.
 //  The hash maps could be rearranged in order not to
 //  include those arguments. It would be an obvious
 //  optimization.

 // TODO:
 //  writing each time interfaces and bodies is a good thing,
 //  but it might be annoyinng for very simple systems or
 //  for some quick testing. It might be interesting to
 //  allow classes to act as their own interfaces, so that
 //  all the methods in the body become part of the dispatcher.

 // TODO:
 //  (not really part of the core): the standard policies are
 //  not terribly efficient, and could be improved. That is
 //  usually not a problem, since the core will cache all
 //  the results, but it would still be nice to review them.

 // TODO:
 //  if the number of compatible methods for a certain prototype
 //  is exactly one, for many policy it is possible to say
 //  that the method applies for *all* subclasses of the parameters
 //  of the prototype. In that case, the hashing mechanism can
 //  be happily replaced with a simple call. The infrastructure
 //  is currently there, but commented (in ClassBuilder).
 //  It could be revived.

 // TODO:
 //  (related to the previous one): for multiple dispatching
 //  it would be more convenient to use a sequence of ifs
 //  rather than a map if the number of compatible methods
 //  is small (less than 8 or so). Adding a general infrastructure
 //  along these lines would be interesting.

 // TODO:
 //  PolyD has no knowledge about generics. A better support
 //  for generics should be incorporated.

 // TODO:
 //  reportedly, not all Java VMs strictly conform to the
 //  same internal names conventions. PolyD currently works
 //  well with Sun's JDK, but it was not tested with different
 //  environments. More testing with different VMs should
 //  be done.

 // QUIRK:
 //  Sun's JDK 1.4 on PowerPC (the MacOS X version) does not
 //  seem inclined to optimize the code generated on-the-fly
 //  by PolyD. The reason is not clear; Sun's JDKs on x86 do
 //  work as expected, generating very efficient code.

 // INVESTIGATE:
 //  subinterfaces and subclasses:
 //  the set of methods that are visible when using
 //  interfaces and bodies is the set of all methods
 //  that are visible and public. A possible
 //  alternative (that would somehow violate the
 //  regular inheritance mechanism in Java) would
 //  be to inspect the classes of the bodies and
 //  recursively extract all public methods declared
 //  in the superclasses, even if they were subsequently
 //  overridden. Not sure it would be wise, but
 //  it might be interesting in certain cases.

 // TODO:
 //  none of the current standard policies allow the user to
 //  decide whether some of the arguments should be dispatched
 //  statically or dynamically. Adding a new policy to do
 //  just that would be simple using raw arguments.

 // INVESTIGATE:
 //  currently the hashmap uses collision chains. That works
 //  well most of the times. However, it still seems a waste
 //  to spend time traversing collision chains when it would
 //  be equally possible to expand dynamically the size of
 //  the hash map in order to accommodate the extra values.
 //  It might be interesting to explore this alternative
 //  implementation.
 
 // ------------------------------


 // Multihash maps use simple parameters up to this value,
 // and arrays beyond that
 static final int maxMethParams=4;
 
 // similar parameter for the factory initializers
 static final int maxFactParams=4;

 // this class stores just the interface
 // and the list of classes of the bodies
 static class DispatcherSignature<T> {
  Class<T> theInterface;
  Class[] theComboClasses;
  public boolean equals(Object o) {
   DispatcherSignature<T>ds=(DispatcherSignature<T>)o;
   if (!theInterface.equals(ds.theInterface)) return false;
   if (theComboClasses.length!=ds.theComboClasses.length) return false;
   for (int i=0;i<theComboClasses.length;i++) {
    if (!theComboClasses[i].equals(ds.theComboClasses[i])) return false;
   }
   return true;
  }
  public int hashCode() {
   int i=theInterface.hashCode();
   for(int j=0;j<theComboClasses.length;j++)
    i=(i*37)+theComboClasses[j].hashCode();
   return i;
  }
  DispatcherSignature(Class<T> theInterface,Class[] comboClasses) {
   this.theInterface=theInterface;
   this.theComboClasses=comboClasses;
  }
 }


 // this class stores the data related to
 // a succesfully built implementation,
 // correspondent to one DispatcherSignature

 private static class DispatcherImplementation<K> {
  List<MethodGroup> groups;
  Class<K> masterClass;
  Factory fact;
  List<Field> selfList;
  boolean specialDispatch;
  DispatcherImplementation(List<MethodGroup> groups,Class<K> masterClass,
              Factory fact,List<Field> selfList,boolean specialDispatch) {
   this.groups=groups;
   this.masterClass=masterClass;
   this.fact=fact;
   this.selfList=selfList;
   this.specialDispatch=specialDispatch;
  }
 }
 
 // I need this code to run under 1.1 -> no java.util.concurrent.ConcurrentHashMap
 // also, the Map will have to be *non* parametric, since Java 1.5 is not able to
 // cope with static parametric fields :( More casts required.
 static java.util.Map autoClassCache=new java.util.HashMap();
 
 // used instead of m.getName()
 // (takes into account renaming)
 static <T> String findName(Descriptor<T> dd,Method m) {  
  if (dd.methInfo.containsKey(m)) {
   Descriptor.MethInfo mi=dd.methInfo.get(m);
   if (mi.theName!=null) return mi.theName;
  }
  return m.getName();
 }


 // used instead of m.getParameterTypes()
 // (takes into account (and excludes) raw arguments
 static <T> Class[] findParameterTypes(Descriptor<T> dd,Method m) {  
  Class[] pars=m.getParameterTypes();
  if (dd.methInfo.containsKey(m)) {
   Descriptor.MethInfo mi=dd.methInfo.get(m);
   return mi.getFiltered(pars);
  }
  return pars;
 }


 // used instead of m.getReturnType()
 // (for future extensions?)
 static <T> Class findReturnType(Descriptor<T> dd,Method m) {  
  return m.getReturnType();
 }


 /**
  * Loads the specified bytecode as a new <code>Class</code>.<p>
  * (Taken almost as-is from the ASM FAQ).
  *
  * @param code  the bytecode, as a <code>byte</code> array
  * @return      the <code>Class</code> corresponding to the loaded bytecode
  */
 private static Class loadThis(String className, byte[] b) {
    //override classDefine (as it is protected) and define the class.
    Class clazz = null;
    try {
      ClassLoader loader = ClassLoader.getSystemClassLoader();
      Class cls = Class.forName("java.lang.ClassLoader");
      java.lang.reflect.Method method =
        cls.getDeclaredMethod("defineClass", new Class[] { String.class, byte[].class, int.class, int.class });

      // protected method invocaton
      method.setAccessible(true);
      Object[] args = new Object[] { className, b, new Integer(0), new Integer(b.length)};
      clazz = (Class) method.invoke(loader, args);
      method.setAccessible(false);
    } catch(Exception e) {
      throw new InternalError(0xAC0,
        "Internal error: PolyD cannot load its own classes using the\n"+
        "system classloader. This shouldn't happen. Probably your VM\n"+
        "prevents Method.setAccessible() from completing successfully.\n"+
        "If you absolutely need setAccessible() to be disabled, please\n"+
        "contact us to obtain a customized version of PolyD that will work."
/*#ifdef14
#else14 */
        ,e
//#endif14
       );
    }
    return clazz;
 }


 
 // use this to presort method arrays, dividing by
 // method name, and then by arity.
 private static class MethodSorter<T> implements java.util.Comparator<Method> {

  Descriptor<T> dd;
  
  public int compare(Method m1,Method m2) {
   String nm1=findName(dd,m1);
   String nm2=findName(dd,m2);
   int nameCompare=nm1.compareTo(nm2);
   if (nameCompare!=0) return nameCompare;
   int ar1=findParameterTypes(dd,m1).length;
   int ar2=findParameterTypes(dd,m2).length;
   if (ar1<ar2) return -1;
   if (ar1>ar2) return 1;
   return 0;
  }
  
  MethodSorter(Descriptor<T> d) {
   dd=d;
  }
 }
 



 
 
 // build a dispatcher given the *annotated* interface and the combos
 // This "build()" parses all the annotations and fills a Descriptor
 // with the acquired information. Although that is not as efficient as
 // parsing the annotations later on, it cleanly separates the Java 5.0
 // dependencies. By building manually a Descriptor object, the user
 // can then use PolyD even without annotations.

 // synchronizing everything might seem overkill. However, we rely on
 // the fact that the "autoClassCache" contains each dispatcher signature
 // after at least one of the combinations is used once. So it should not
 // be slow on the long term. I still would love to use
 // ConcurrentHashMap or similar stuff, but that's not an option here.

/*#ifdef14
#else14 */
 
 
 
 
 /** 
  * Creates a new dispatcher.<p>
  * This is the main method of PolyD, used to build a dispatcher using
  * the annotation-based mechanism. Just define the desired interface and
  * implementing bodies, and use this method to build your new dispatcher.
  * Normally, this is the only method of this class that you will require.<p>
  * This method is not available in the pre-5.0 version of PolyD.
  * @param interf the interface that defines the available messages
  * @param bodies the list of objects belonging to the implementing classes
  * @return the new dispatcher, which implements the supplied interface
  */
 public synchronized static <T,K extends T> T build(Class<T> interf,Object... bodies) {

  Class[] comboClasses=new Class[bodies.length];
  for (int i=0;i<bodies.length;i++) {
   comboClasses[i]=bodies[i].getClass();
  }

  DispatcherSignature<T> ds=new DispatcherSignature<T>(interf,comboClasses);
  DispatcherImplementation<K> di;
  if (!autoClassCache.containsKey(ds)) { // it *is not* in the autocache

   Descriptor<T> dd = parseAnnotations(interf, comboClasses, ds);
   di=buildImplementation(dd,false);
   autoClassCache.put(ds,di);
  }
  else { // it *is* in the cache
   di=(DispatcherImplementation<K>)(autoClassCache.get(ds));
  }
  
 return buildNewDispatcher(ds,di,bodies);
}

 /**
  * Creates a new dispatcher from a registered <code>Descriptor</code>.
  * This method could be removed from future versions of PolyD,
  * and its use is not recommended.<p>
  * This method is not available in the pre-5.0 version of PolyD.
  * @deprecated Use {@link #build} instead.

  * @param interf the interface used to build the new dispatcher
  * @param body the class of the single body (that must implement <code>interf</code>)
  * @param args the arguments used to create a new instance of <code>body</code>
  * @param argsClasses the classes of the arguments used to create the new instance
  * @return the new dispatcher, which implements the supplied interface
  */
 public synchronized static <T,K extends T> K buildSpecial(Class<T> interf, Class<K> body, Object[] args, Class[] argsClasses) {
  Class[] comboClasses=new Class[1];
  comboClasses[0]=body;
  DispatcherSignature<T> ds=new DispatcherSignature<T>(interf,comboClasses);
  DispatcherImplementation<K> di;
  if (!autoClassCache.containsKey(ds)) { // it *is not* in the autocache

   Descriptor<T> dd = parseAnnotations(interf, comboClasses, ds);
   di=buildImplementation(dd,true);
   autoClassCache.put(ds,di);
  }
  else { // it *is* in the cache
   di=(DispatcherImplementation<K>)(autoClassCache.get(ds));
  }
  return buildSpecialNewDispatcher(ds,di,args,argsClasses);
 }

 private static <T> Descriptor<T> parseAnnotations(Class<T> t, Class[] comboClasses, DispatcherSignature<T> ds) {
  
  // Let's accumulate info from the annotations.
  Descriptor<T> dd=new Descriptor<T>(ds);
  // check for the @PolyD tag
  if (!t.isAnnotationPresent(ovm.polyd.tags.PolyD.class)) {
   throw new DispatcherCreationException(0x101,
    "Interface \""+t.getName()+"\" does not look like\n"+
    "  a PolyD descriptor (the @PolyD tag is missing)."
   );
  }
  if (t.isAnnotationPresent(DispatchingPolicy.class)) {
   dd.setDispatching(t.getAnnotation(DispatchingPolicy.class).value());
  }
  if (t.isAnnotationPresent(InvocationPolicy.class)) {
   dd.setInvocation(t.getAnnotation(InvocationPolicy.class).value());
  }
  if (t.isAnnotationPresent(OnMissing.class)) {
   dd.setMissingHandling(t.getAnnotation(OnMissing.class).value());
  }
  for (Method mt : t.getMethods()) {
   
   boolean methodInfoPresent=false;

   // let's scan for annotations. Is there a custom policy?
   if (mt.isAnnotationPresent(DispatchingPolicy.class)) {
    dd.setMethodDispatching(mt,mt.getAnnotation(DispatchingPolicy.class).value());
   }
   
   // Since we are scanning the protos, check for the invocation policy
   if (mt.isAnnotationPresent(InvocationPolicy.class)) {
    dd.setMethodInvocation(mt,mt.getAnnotation(InvocationPolicy.class).value());
   }
   
   // And also, since we're scanning the protos, let's fix the missing handling.
   if (mt.isAnnotationPresent(OnMissing.class)) {
    dd.setMethodMissingHandling(mt,mt.getAnnotation(OnMissing.class).value());
   }

   if (mt.isAnnotationPresent(Name.class)) {
    dd.setMethodName(mt,mt.getAnnotation(Name.class).value());
   }

   Annotation avv[][]=mt.getParameterAnnotations();
   Class[] pars=mt.getParameterTypes();

   // Since we are scanning the protos, check whether there are raw parameters,
   boolean[] raws=new boolean[avv.length];
   boolean rawUsed=false;
   for (int ix=0;ix<avv.length;ix++) {
    for (Annotation a : avv[ix]) {
     if (a instanceof Raw) {
      rawUsed=raws[ix]=true;
     }
    }
   }
   if (rawUsed) dd.setMethodRawClasses(mt,raws);

   // Since we are scanning the protos, check whether there are forced classes
   Class[] asClasses=new Class[avv.length];
   boolean asIsUsed=false;
   for (int ix=0;ix<avv.length;ix++) {
    for (Annotation a : avv[ix]) {
     if (a instanceof As) {
      asIsUsed=true;
      asClasses[ix]=((As)a).value();
     }
    }
   }
   if (asIsUsed) dd.setMethodAsClasses(mt,asClasses);

   // Since we are scanning the protos, check whether there are defaults for nulls,
   // BUT ignore if there already is a forcedClass
   // (avv and pars are already ready)
   // Annotations restrict us to Classes, but using the API the
   // values can be arbitrary.
   Class[] nullDefaults=new Class[avv.length];
   boolean nullDefaultsUsed=false;
   for (int ix=0;ix<avv.length;ix++) {
    if (asClasses[ix]==null) // @As prevails on @Null
    for (Annotation a : avv[ix]) {
     if (a instanceof IfNull) {
      nullDefaultsUsed=true;
      nullDefaults[ix]=((IfNull)a).value();
     }
    }
   }
   if (nullDefaultsUsed) dd.setMethodNullDefaults(mt,nullDefaults);

   // Since we are scanning the protos, inspect the @Preload annotations, so that the
   // maps can be preinitialized at creation-time.
   // it is not strictly necessary, but the @Preload annotations are converted to
   // an array of array of classes. That is useful if annotations are actually not
   // used, but the MethodGroup is initialized in some other way (so that the features
   // can be used even if proto and bodies are written in pre-5.0 Java
   // Annotations restrict us to Classes, but using the API the
   // values can be arbitrary.
   Class[][] preload=null;
   if (mt.isAnnotationPresent(Preload.class)) {
    Seq[] preloadList=mt.getAnnotation(Preload.class).value();
    int size=preloadList.length;
    preload=new Class[size][];
    for (int iy=0;iy<size;iy++) {
     preload[iy]=preloadList[iy].value();
    }
    dd.setMethodPreload(mt,preload);
   }

  }
  for (Class comboClass : comboClasses) {
   for (java.lang.reflect.Field f : comboClass.getFields()) {
    if (f.isAnnotationPresent(Self.class)) {
     if (f.getType().equals(t)) {
      dd.setSelfField(f);
     }
    }
   }
   // There could also be @Name tags in the combos (the only two allowed tags
   // in combos are @Self and @Name, but I ignore the others)
   for (Method mt : comboClass.getMethods()) {
    if (mt.isAnnotationPresent(Name.class)) {
     dd.setMethodName(mt,mt.getAnnotation(Name.class).value());
    }
   }
   
   }
  return dd;
 }

//#endif14
 
 
 // package visibility
 synchronized static <T,K extends T> Factory register(Descriptor<T> dd,boolean specialDispatch) {
  DispatcherImplementation<K> di=(DispatcherImplementation<K>)buildImplementation(dd,specialDispatch);
  autoClassCache.put(dd.theSignature,di);
  return di.fact;
 }

 // build a dispatcher given the Descriptor (build using interface
 // and other info) and the combos. It is to be called only when
 // building descriptors manually.
 // Since it is not called by the build() above, more checks are
 // made.
 /** 
  * Creates a new dispatcher from a registered <code>Descriptor</code>.<p>
  * This method is not normally necessary, and it is mainly of use in
  * the pre-5.0 version of PolyD. The preferred approach, in
  * that case, is building a <code>Descriptor</code> and then using
  * a <code>Factory</code>. However, sometimes it might not be
  * practical to deal with factories in the client code.<p>
  * In those cases, after a <code>Descriptor</code> has been registered this
  * method can be used to generate new matching dispatchers
  * without using a <code>Factory</code>.
  * @see Descriptor
  * @see Factory
  * @param interf the interface that defines the available messages
  * @param bodies the list of objects belonging to the implementing classes
  * @return the new dispatcher, which implements the supplied interface
  */
 public synchronized static <T,K extends T> T buildFromDescriptor(Class<T> interf,Object... bodies) {

  Class[] comboClasses=new Class[bodies.length];
  for (int i=0;i<bodies.length;i++) {
   comboClasses[i]=bodies[i].getClass();
  }

  DispatcherSignature<T> ds=new DispatcherSignature<T>(interf,comboClasses);
  DispatcherImplementation<K> di=null;
   if (autoClassCache.containsKey(ds)) // it *is* in the autocache
    di=(DispatcherImplementation<K>)(autoClassCache.get(ds));
   else {
    throw new DispatcherCreationException(0x102,
     "I do not know of any descriptor which matches these interface and bodies.\n"+
     "If you have one, please use register() to install it before using\n"+
     "buildFromDescriptor()."
    );
   }
   return buildNewDispatcher(ds,di,bodies);
 }
 
 /**
  * Creates a special dispatcher from a registered <code>Descriptor</code>.
  * A new instance of <code>body</code> will be created without passing
  * any arguments to the constructor. This method could be removed from
  * future versions of PolyD, and its use is not recommended.
  * @deprecated Use {@link #buildFromDescriptor} instead.

  * @param interf the interface used to build the new dispatcher
  * @param body the class of the single body (that must implement <code>interf</code>)
  * @return the new dispatcher, which implements the supplied interface
  */
 public static <T,K extends T> K buildSpecialFromDescriptor(Class<T> interf, Class<K> body) {
  return buildSpecialFromDescriptor(interf,body,new Object[]{},new Class[]{});
 }

 /**
  * Creates a special dispatcher from a registered <code>Descriptor</code>.
  * This method could be removed from future versions of PolyD,
  * and its use is not recommended.
  * @deprecated Use {@link #buildFromDescriptor} instead.

  * @param interf the interface used to build the new dispatcher
  * @param body the class of the single body (that must implement <code>interf</code>)
  * @param args the arguments used to create a new instance of <code>body</code>
  * @param argsClasses the classes of the arguments used to create the new instance
  * @return the new dispatcher, which implements the supplied interface
  */
  public synchronized static <T,K extends T> K buildSpecialFromDescriptor(Class<T> interf, Class<K> body, Object[] args, Class[] argsClasses) {

  DispatcherSignature<T> ds=new DispatcherSignature<T>(interf,new Class[] {body});
  DispatcherImplementation<K> di=null;
   if (autoClassCache.containsKey(ds)) // it *is* in the autocache
    di=(DispatcherImplementation<K>)(autoClassCache.get(ds));
   else {
    throw new DispatcherCreationException(0x103,
     "I do not know of any descriptor which matches these interface and bodies.\n"+
     "If you have one, please use register() to install it before using\n"+
     "buildFromDescriptor()."
    );
   }
   return buildSpecialNewDispatcher(ds,di,args,argsClasses);
 }
 
// the groups, during this initial partitioning, are kept
// in a List of SigGroup (signature groups)

  private static class SigGroup {
   Method[] protos;
   Method[] bodies;
  }

 
 // build a dispatcher given the Descriptor (build using interface
 // and other info). Private to ensure consistency.
 private static <T,K extends T> DispatcherImplementation<K> buildImplementation(Descriptor<T> dd,boolean specialDispatch) {

  //
  // comboClasses is used in the code generation. The actual objects
  // must *NOT* be used before the actual object
  // instantiation
  //
  DispatcherSignature<T> ds=dd.theSignature;
  Class[] comboClasses=ds.theComboClasses;
  Class <T> theInterface=ds.theInterface;

  // verify that the interface is an interface for real.
  //
  // NEW:
  //  There are now two possibilities:
  //  1) theInterface is an interface, and I have a list of implementing classes/instances
  //  2) theInterface is a class, and I have no additional bodies specified.
  //       in this second case theInterface is the only body and its own interface.

  if (!theInterface.isInterface()) {
   throw new DispatcherCreationException(0x104,
    "The base you specified to build a new dispatcher is not an interface."
   );
  }
  if (specialDispatch) {
   //
   // make sure specialDispatch can really be used.
   //
   if (comboClasses.length!=1) {
    throw new DispatcherCreationException(0x105,
     "Special dispatch can only be used with exactly one implementing class"
    );
   }
   if (!theInterface.isAssignableFrom(comboClasses[0])) {
    throw new DispatcherCreationException(0x106,
     "Special dispatch can only be used if the supplied class implements the interface"
    );
   }
  }


  // inspect the prototypes in the interface
  Method[] protos=theInterface.getMethods();
  if (protos.length==0) {
   throw new DispatcherCreationException(0x107,
    "First argument to method() declares no prototypes"
   );
  }
  for (Method mt : protos) {
    int mo=mt.getModifiers();
    if ( (mo & java.lang.reflect.Modifier.STATIC)!=0) {
     // should be impossible, but who knows?
     throw new DispatcherCreationException(0x108,
      "Method "+Utils.methodString(mt)+
      " cannot be static in the definition of the interface."
     );
    }
  }
  

/*

  Let's try to match the prototypes with the bodies.
  The idea is simple, sort the prototypes, then divide them
  in chunks according to name/arity.
  Then sort in the same way the bodies, and scan the list
  of sorted bodies in parallel with the list of sorted
  and partitoned prototypes, so that the bodies that
  match are put in the proper group.

  The simple idea is unfortunately a bit tricky to implement,
  at least trying to use mostly arrays as I did. The resulting
  code is not nice, but it works. I prefer not to rewrite it
  in a better way, lest I end up breaking something.

*/



  
  MethodSorter<T> ms=new MethodSorter<T>(dd);
  java.util.Arrays.sort(protos, ms);
  int i;
  // divide them in groups
  // using this list (or some other more efficient mechanism
  // I'll think about in some future)
  LinkedList<SigGroup> sgroups=new LinkedList<SigGroup>();
  SigGroup mg=new SigGroup();
  // Scan the sorted array, and find which are the
  // positions at which arity or method name change.
  // Split into multiple arrays accordingly.
  int mgarity=findParameterTypes(dd,protos[0]).length;
  String mgname=findName(dd,protos[0]);
  sgroups.add(mg);
  int last=0;
  for (i=1;i<protos.length;i++) {
   int newar=findParameterTypes(dd,protos[i]).length;
   String newname=findName(dd,protos[i]);
   if (mgarity!=newar || !mgname.equals(newname)) {
    mg.protos=new Method[i-last];
    System.arraycopy(protos,last,mg.protos,0,i-last);
    mg=new SigGroup();
    mgarity=newar;
    mgname=newname;
    sgroups.add(mg);
    last=i;
   }
  }
  mg.protos=new Method[i-last];
  System.arraycopy(protos,last,mg.protos,0,i-last);


  // The prototypes are now known and divided by name/arity,
  // each SigGroup stores the protos with same name/arity
  // (according to the renaming) and has the following fields
  // initialized: arity, name, master class (owner)
  // (besides the set of prototypes)


  // 
  // do not check for duplicates in the interface!
  // since we have renaming, duplicates may be present
  //

  //
  // also, do not check for redundant prototypes.
  // They might be using different policies, and
  // additionally all protos must be implemented by
  // the master class
  //
   
    
  /*
   * The prototypes are now known and divided by name/arity. The same process
   * should now be applied to the bodies contained in the combos. First of all
   * the methods should be extracted from the combos, then concatenated
   * together. Then sorted, then split as above. For each group, it should be
   * verified that the list of bodies is consistent, and that each of them
   * really fits one of the available prototypes.
   */

  
  Method[][] allBodies=new Method[comboClasses.length][];

  int nBodies=0;
  for (i=0;i<comboClasses.length; i++) {
   allBodies[i]=comboClasses[i].getMethods();
   nBodies+=allBodies[i].length;
  }
  Method[] sortedBodies=new Method[nBodies];
  nBodies=0;
  for (i=0;i<comboClasses.length; i++) {
   System.arraycopy(allBodies[i],0,sortedBodies,nBodies,allBodies[i].length);
   nBodies+=allBodies[i].length;
  }
  allBodies=null; // GC
  java.util.Arrays.sort(sortedBodies, ms);
 

  
  // at this stage we are only concerned with arity and name,
  // we do not investigate the specific parameter types. That is
  // also consistent with the fact that the sorting was made by
  // MethodSorter.

  // I *could* use getParameterTypes, rather than findParameterTypes,
  // since @Raw only appears in protos, but it's best to use a proper
  // infrastructure, just in case that should change
  
  i=0; // progresses over sortedBodies, first index
  for (SigGroup m : sgroups) {
   int marity=findParameterTypes(dd,m.protos[0]).length;
   String mname =findName(dd,m.protos[0]);
   while (i<sortedBodies.length && !(findName(dd,sortedBodies[i]).equals(mname) &&
     marity==findParameterTypes(dd,sortedBodies[i]).length)) {
    // If we are here it means that there is a mismatch between protos
    // and bodies. if i==sortedBodies.length there are no bodies, so
    // there is a proto without body. Otherwise it is not certain
    // whether there is a proto without body or a body without a proto.

    if (existsSimilarMethod(dd,sortedBodies[i],protos)) {
     // the body has a proto. A proto does not have a body -> error
     throw new DispatcherCreationException(0x109,
      "Message \""+mname+"\", arity "+
       marity+" has no implementing methods."
     );
    } else {
     // the body does not have a proto. The proto has a body.
     // The body is a supporting method, it is removed from the list.
//     System.err.println("Warning: method
//          = "+methodString(sortedBodies[i])+
//              " does not match any prototype.");
     i++;
    }
   } 
   int start=i;
   while (i<sortedBodies.length && findName(dd,sortedBodies[i]).equals(mname) &&
           marity==findParameterTypes(dd,sortedBodies[i]).length) i++;
   m.bodies=new Method[i-start];
   System.arraycopy(sortedBodies,start,m.bodies,0,i-start);
  }
  sortedBodies=null; // GC
  protos=null; // GC


  ///////////////////////////
  
  /*
   * The first approximate checks are passed. Now the rules regarding
   * parameters/arguments and return types must be satisfied.
   *  
   */
  
  boolean troubles=false;
  String troublesString="";
  for (SigGroup g : sgroups) {
   int blen=g.bodies.length,bi;
   int plen=g.protos.length,pi;
   boolean bodyUsed[]=new boolean[blen];
   boolean protoSatisfied[]=new boolean[plen];

   // cache protos parameters and return type
   Class protoParams[][]=new Class[plen][];
   Class protoReturn[]=new Class[plen];
   for (pi=0;pi<plen;pi++) {
    protoParams[pi]=findParameterTypes(dd,g.protos[pi]);
    protoReturn[pi]=findReturnType(dd,g.protos[pi]);
   }
   
   for (bi=0;bi<blen;bi++) {
    Class bodyParams[]=findParameterTypes(dd,g.bodies[bi]);
    Class bodyReturn=findReturnType(dd,g.bodies[bi]);
    for (pi=0;pi<plen;pi++) {
     Diff how=Utils.classArrayDiff(protoParams[pi],bodyParams);
          // if NotCompatible, do nothing
     if (how!=NotCompatible) {
      // Must check return types (which could be void or primitive)
      if (!Utils.isEqualOrSub(bodyReturn,protoReturn[pi])) {
       if (troubles) troublesString+="\n";
       troublesString+="Method "+
         Utils.methodString(g.bodies[bi])+
         " can be called through the prototype "+
         Utils.methodString(g.protos[pi])+
         " but the return type "+bodyReturn.getName()+
         " is not equal or a subclass of "+
         protoReturn[pi].getName()+
         ", as required by the prototype.";
       troubles=true;
      }
      // body is used,
      bodyUsed[bi]=true;
      if (how==Equal || how==Sub) {
       // body is more generic, covers the proto.
       protoSatisfied[pi]=true;
      } // else if (how==Super || how==Crossed) {}
     }
    }    
   }
   
   for (bi=0;bi<blen;bi++) {
    if (!bodyUsed[bi]) {
     if (troubles) troublesString+="\n";
     troublesString+="Method "+Utils.methodString(g.bodies[bi])+
      " is not compatible with any prototype defined in interface \""+
      theInterface.getName()+"\".";
     troubles=true;
    }
   }
   for (pi=0;pi<plen;pi++) {
    if (!protoSatisfied[pi]) {
     if (troubles) troublesString+="\n";
     troublesString+="Message "+Utils.methodString(g.protos[pi])+
      " has no compatible implementing body.";
     troubles=true;
    }
   }
  }
  if (troubles)
   throw new DispatcherCreationException(0x110,troublesString);

  
  
  /*
   * Up to this point, each SigGroup groups multiple protos
   * and a MethInfo might or might not be present for a
   * certain prototype.
   *
   * We now split each SigGroup into multiple MethodGroups
   * so that each MethodGroup refers to a single prototype
   * and a MethInfo is present in any case.
   */

  LinkedList<MethodGroup> groups=new LinkedList<MethodGroup>();
  for (SigGroup g : sgroups) {
   for (i=g.protos.length-1;i>=0;i--)
    // I will look for the matching MethInfo in dd
    groups.add(new MethodGroup(g.protos[i],g.bodies,dd));
  }




  /*
   * At this stage it is possible to check consistency according to the
   * as classes, verify consistency and finally replace the set of
   * bodies in each MethodGroup with the compatibleSet.
   *
   */

  
  for (MethodGroup g : groups) {

     // we now know which are the forced classes (@As), hence we have a *new* proto that needs to be checked,
     // and covered by at least one available body. We might have checked just this new combination
     // rather than the previous one, but the @As clauses are ignored by overloading, for instance.
     // So both have to be checked.
     //
     // The combination that we need to verify as covered by at least one body is the same sequence of
     // parameters as the proto, but with some classes changed according to forcedClasses.
     //
      Boolean good=false;
      int bi,blen=g.bodies.length;
      Class[] forcedPars=g.getForcedFiltered();

      for(bi=0;bi<blen;bi++) {
       Diff how=Utils.classArrayDiff(forcedPars,findParameterTypes(dd,g.bodies[bi]));
       if (how==Equal || how==Sub) {
        //  body is more generic, covers the proto.
        good=true; break;
       }
      }
      if (!good) {
       throw new DispatcherCreationException(0x111,
        "There is no implementation that can be called through the prototype\n"+
        g.proto.toString()+"\n"+
        " (in "+theInterface.getName()+"), as forced by the @As clauses."
       );
      }

   
   if (g.protoInfo.theDispatching==null) {
    throw new DispatcherCreationException(0x112,
     "No dispatching policy is associated to prototype "+g.proto.toString()+"\n"+
     "Please add a @DispatchPolicy tag to the prototype or to the interface."
    );
   }

   if (specialDispatch && g.protoInfo.theInvocation!=null) {
    throw new DispatcherCreationException(0x113,
     "Special dispatching is incompatible with custom invocation policies."
    );
   }
  
  }
  /* ................ */  
  

  
   // replace the list of all bodies having the same name/arity of this proto
   // with the list of bodies which are strictly compatible according to each
   // dispatching policy
   // and/or perform a consistency check on the set of methods.

   for (MethodGroup g : groups) {
    g.bodies=g.compatibleSet();
    if (g.bodies.length==0) {
     throw new DispatcherCreationException(0x115,
      "Prototype: "+g.proto.toString()+"\n"+
      "has no matching body according to dispatching policy "+
          g.protoInfo.theDispatching.toString()+"."
     );
    }
   }
  
   String[] comboMangled = mangleClasses(comboClasses);

  //
  // the name of the master class.
  //
  String className="A"+Uniq.uniq();

   /*
    * Generation of the master class.
    */
   Class<K> masterClass=null;

   byte[] code=ClassBuilder.createMasterCode(groups,className,theInterface,
                            comboClasses,comboMangled,specialDispatch);
   // Extremely useful for debug:
   // ClassBuilder.writeCode(code,className);

   masterClass=loadThis("ovm.polyd.PolyD$"+className,code);
   
   try {
    java.lang.reflect.Method mm=masterClass.getDeclaredMethod("SPECIAL$$STATIC$$INIT",new Class[]{List.class});
    mm.invoke(null,new Object[]{groups});
   } catch(Throwable e) {
/*#ifdef14
    throw new InternalError(0x1C0,"Internal error while calling static initializer");
#else14 */
    if (e instanceof java.lang.reflect.InvocationTargetException)
     e=((java.lang.reflect.InvocationTargetException)e).getTargetException();
    throw new InternalError(0x1C0,"Internal error while calling static initializer",e);
//#endif14
    
   }
 
   // ----------------------------------------------------------------------
   //   

// if !specialDispatch, I use a Factory for dispatcher creation
   
   Factory fact=null;
   if (!specialDispatch) {
    byte[] factoryCode=ClassBuilder.createFactoryCode(groups,className,theInterface,comboClasses,comboMangled,dd.selfList);
    // ClassBuilder.writeCode(factoryCode,"FACT"+className);

    Class<Factory> factoryClass=loadThis("ovm.polyd.Factory$"+className,factoryCode);
    try {
     fact=(Factory)(factoryClass.newInstance());
    } catch(Exception e) {
     throw new InternalError(0x1C3,"Internal error while creating a new Factory"
/*#ifdef14
#else14 */
     ,e
//#endif14
     );
    }
   }
   
   DispatcherImplementation<K> di=new DispatcherImplementation<K>(groups,masterClass,fact,dd.selfList,specialDispatch);

   return di;
 }


 private static <T,K extends T> T buildNewDispatcher(DispatcherSignature<T> ds,
                                        DispatcherImplementation<K> di,Object[] combos) {
  if (di.specialDispatch) {
   throw new DispatcherCreationException(0x116,
    "The descriptor was registered as a special dispatch.\n"+
    "Please use buildSpecial() or buildSpecialFromDescriptor() instead."
   );
  }
  List<Field> selfList=di.selfList;
  Class<T> theInterface=ds.theInterface;
  // we do not have specialDispatch -> use the factory
  // it looks somewhat inefficient here, it would arguably be
  // better if multiple, arity-dependant versions of this piece of code
  // were created. In any case, it is quite efficient when using
  // directly the factory to create dispatchers. Better than using
  // reflection, in any case.
  Factory fact=di.fact;
  K newMaster;
  switch (combos.length) {
   case 1: newMaster=(K)(fact.getDispatcher1(combos[0]));
   break;
   case 2: newMaster=(K)(fact.getDispatcher2(combos[0],combos[1]));
   break;
   case 3: newMaster=(K)(fact.getDispatcher3(combos[0],combos[1],combos[2]));
   break;
   case 4: newMaster=(K)(fact.getDispatcher4(combos[0],combos[1],combos[2],combos[3]));
   break;
   default: newMaster=(K)(fact.getDispatcherN(combos));
  }
// Self fields are set up by the factory
  return newMaster;
 }

 // only used for special dispatching (no combos/arguments for construction supplied)
 private static <T, K extends T> K buildSpecialNewDispatcher(
   DispatcherSignature<T> ds, DispatcherImplementation<K> di, Object[] args, Class[] argsClasses) {
  if (!di.specialDispatch) {
   throw new DispatcherCreationException(0x117,
    "The descriptor was not registered as a special dispatch.\n"+
    "Please use build() or buildFromDescriptor() instead."
   );
  }
  List<Field> selfList = di.selfList;
  Class<T> theInterface = ds.theInterface;
  Class<K> masterClass = di.masterClass;
  K newMaster = null;
  try {
   java.lang.reflect.Constructor<K> k = masterClass
     .getConstructor(argsClasses);
   newMaster = k.newInstance(args);
  } catch (Exception e) {
   throw new DispatcherCreationException(0x118,
    "Special dispatching construction failed.\n"+
    "The arguments you supplied do not match any constructor in "+ds.theComboClasses[0].getName()+",\n"+
    "or the constructor is not declared as public."
   );
  }
  for (Field f : selfList) {
   if (f.getDeclaringClass().isAssignableFrom(masterClass)) {
    setSelf(f, newMaster, theInterface, newMaster);
   }
  }
  return newMaster;
 }


 // IGNORES fileds which were not declared with the exact same class of
 // the main interface we are using. Thanks to that, a single body can be shared
 // by multiple built dispatchers, as long as they have rely on distinct
 // interfaces
 private static <T,K extends T> void setSelf(java.lang.reflect.Field f, Object combo, Class<T> t, K newMaster) {
  if (!f.getType().equals(t)) {
   return;
  }     
  try {
   f.set(combo,newMaster);
  } catch (Exception e) {
   throw new DispatcherCreationException(0x119,
    "@Self field \""+f.getName()+"\", declared in "+combo.getClass().getName()
     +", could not be initialized (not accessible?)."
   );
  }
 }
 
 
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////
 
 //
 // checks the name/arity, looking for a Method similar to "test"
 //
 static private <T> boolean existsSimilarMethod(Descriptor<T> dd, Method test,Method[] meth) {
  
  String name=findName(dd,test);
  Class args[]=findParameterTypes(dd,test);
  
  if (meth.length==0) return false;
  
  int ar=args.length; // arity
  
  for (Method m : meth) {
   if (m!=null &&
     findName(dd,m).equals(name) &&
     ar==findParameterTypes(dd,m).length) {
       return true;
   }
  }
  
  // no match.

  return false;
 }


////////////////////////

// give an estimate for the size of a lookup table 
 
 static int guessTableSize(MethodGroup methGroup_) {
  int size;
  int hint=methGroup_.bodies.length*3;
  size = 128;
  while ( size < hint ) size <<= 1;
  return size;
 }

 private static String[] mangleClasses(Class[] comboClasses) {
  String[] comboMangled=new String[comboClasses.length];
  for (int i=0;i<comboClasses.length;i++) {
   comboMangled[i]="instance_"+comboClasses[i].getName().replace('.','_')+"_"+Uniq.uniq();
  }
  return comboMangled;
 }

 
 // This class is similar to java.util.concurrent.atomic.AtomicInteger,
 // but works with pre-5.0 code as well
 private static class Uniq {
  static private int i=0;
  synchronized static public int uniq() {
   return i++;
  }
 }

}
