/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd;
import java.lang.reflect.Method;
import static ovm.polyd.Utils.Diff.*;

/**
 * A convenience class that implements
 * a few handy utilities.<p>
 * 
 * @author Antonio Cunei
 * @version $Revision: 1.1 $, $Date: 2005/04/02 05:59:05 $
 * @see Utils.Diff
 */
abstract public class Utils {
 private Utils() {}

/*#ifdef14
    public static final class Diff
    {

        public static final Diff Sub;
        public static final Diff Equal;
        public static final Diff Super;
        public static final Diff Crossed;
        public static final Diff NotCompatible;
        private String name;

        public String toString()
        {
            return (name);
        }

        static
        {
            Sub = new Diff("Sub");
            Equal = new Diff("Equal");
            Super = new Diff("Super");
            Crossed = new Diff("Crossed");
            NotCompatible = new Diff("NotCompatible");
        }

        private Diff(String s)
        {
            name=s;
        }
    }
#else14 */
 /**
  * Used to describe the result of a comparison between two
  * classes or arrays of classes.
  * @author cuneia
  * @version $Revision: 1.1 $, $Date: 2005/04/02 05:59:05 $
  * @see Utils
  */
  static public enum Diff {
   /**
    * The first argument is a strict subclass of the second.
    * If arrays are compared, all arguments of the first array
    * are equal or subclasses of the second, but they are not
    * all identical.
    */
   Sub,
   /**
    * The first argument is identical to the second.
    */
   Equal,
   /**
    * The first argument is a strict superclass of the second.
    * If arrays are compared, all arguments of the first array
    * are equal or superclasses of the second, but they are not
    * all identical.
    */
   Super,
   /**
    * When comparing two arrays of classes, for some of the
    * positions the first class is a strict subclass of the second while
    * for other positions the first class is a strict superclass of the
    * second. For all of the positions, the corresponding pairs of classes
    * are related (either equal, or sub, or super).
    */
   Crossed,
   /**
    * The two classes are not related, meaning that none of the
    * two is equal or a subclass of the other. When comparing arrays of
    * classes, at least for one of the positions the corresponding
    * classes are not related.
    */
   NotCompatible }
//#endif14

 /**
  * Compares two classes.
  * @param a the first class that will be compared
  * @param b the second class that will be compared
  * @return the result of the comparison
  */
  static public Diff classDiff(Class a,Class b) {
  if (a==b) return Equal;
  if (((Class<?>)a).isAssignableFrom(b)) return Super;
  if (((Class<?>)b).isAssignableFrom(a)) return Sub;
  return NotCompatible;
 }

 /**
  * Compares two arrays of classes.
  * @param a the first array that will be compared
  * @param b the second array that will be compared
  * @return the result of the comparison
  */
  static public Diff classArrayDiff(Class[] a,Class[] b) {
  int len=a.length;
  int i;
  Diff direction=Equal;
  for (i=0;i<len;i++) {
   Diff now=classDiff(a[i],b[i]);
   if (now==NotCompatible) { return NotCompatible; }
 
   if (direction==Equal) {
    if (now==Super) { direction=Super; continue; }
    if (now==Sub)   { direction=Sub; continue; }
    // if Equal
    continue;
   }
   if (direction==Super) {
    if (now==Sub)   { direction=Crossed; continue; }
    // if Super or Equal
    continue;
   }
   if (direction==Sub) {
    if (now==Super) { direction=Crossed; continue; }
    // if Sub or Equal
    continue;
   }
   // if Crossed, it remains Crossed unless NotCompatible
  }
  return direction;
 }

  
  
 /**
  * Compares two arrays of classes, and attempts to build an
  * array that is a subclass of both arrays.<p>
  * If the result of the method is anything but <code>NotCompatible</code>,
  * the two arrays of classes will be combined and the
  * resulting combination stored in the array specified by
  * the third argument.<p>
  * The new array will contain the narrowest element of each
  * pair (the class which is a subclass of the other one).<p>
  * If the result of the method is <code>NotCompatible</code>,
  * the data the third array will receive is unspecified.
  * @param a the first array that will be compared
  * @param b the second array that will be compared
  * @param cross the array where the new combination will be stored
  * @return the result of the comparison
  */ 
 // there are a few redundant "continue", but it's clearer
 // this way.
 // the third argument must point to a Class[] that will receive
 // the common sub-tuple in case the result is "Crossed", and
 // the lower sub-tuple if the result is anything but "notCompatible"
 static public Diff classArrayDiff(Class[] a,Class[] b,Class[] cross) {
  int len=a.length;
  int i;
  Diff direction=Equal;
  for (i=0;i<len;i++) {
   Diff now=classDiff(a[i],b[i]);
   if (now==NotCompatible) { return NotCompatible; }
 
   if (now==Super)
    cross[i]=b[i];
   else
    cross[i]=a[i];
 
   if (direction==Equal) {
    if (now==Super) { direction=Super; continue; }
    if (now==Sub)   { direction=Sub; continue; }
    // if Equal
    continue;
   }
   if (direction==Super) {
    if (now==Sub)   { direction=Crossed; continue; }
    // if Super or Equal
    continue;
   }
   if (direction==Sub) {
    if (now==Super) { direction=Crossed; continue; }
    // if Sub or Equal
    continue;
   }
   // if Crossed, it remains Crossed unless NotCompatible
  }
  return direction;
 }

 /**
  * Compares two classes, and determines whether the
  * first one is equal or a subclass or the second.
  * @param a the first class that will be compared
  * @param b the second class that will be compared
  * @return the result of the comparison
  */
 static public boolean isEqualOrSub(Class a,Class b) {
  return (((Class<?>)b).isAssignableFrom(a));
 }
 
 
 /**
  * Compares two arrays of classes, and determines
  * whether each element of the first is equal or a
  * subclass of the second.<p>
  * The first array can contain null values, which
  * will not be included in the overall comparison.
  * @param a the first array that will be compared
  * @param b the second array that will be compared
  * @return the result of the comparison
  */
 static public boolean isEqualOrSubArrayNull(Class[] a,Class[] b) {
  int len=a.length;
  for (int i=0;i<len;i++) {
   if (a[i]==null) continue;
   if (!isEqualOrSub(a[i],b[i])) return false;
  }
  return true;
 }

 /**
  * Determines whether any of those methods has
  * has a given combination of formal parameters.<p>
  * @param params an array of desired parameter classes
  * @param meth an array of methods
  * @return the index of the first method that matches, if
  * one is found, or -1 if not found.
  */
 // looks for this list of parameters in the given methods
 static public int isThere(Class[] params,Method[] meth) {
  int ar=params.length;
  int ml=meth.length;
  int k,q;
outer:
  for (k=0;k<ml;k++) {
   Class[] test=meth[k].getParameterTypes();
   for (q=0;q<ar;q++) {
    if (classDiff(test[q],params[q])!=Equal) {
     continue outer;
    }
   }
   return ml;
  }
  return -1;
 }

 /**
   * Generates a user-friendly string representation of a method.
   * 
   * @param m the method than needs to be printed
   * @return the string representation of the method
   */
  public static String methodString(Method m) {
   String s=m.getName()+"(";
   String sep="";
   for (Class c : m.getParameterTypes()) {
    s+=sep+c.getName();
    sep=",";
   }
   s+="), defined in "+m.getDeclaringClass().getName();
/*#ifdef14
#else14 */
   if (m.isAnnotationPresent(ovm.polyd.tags.Name.class))
    s+=" (renamed as "+m.getAnnotation(ovm.polyd.tags.Name.class).value()+")";
//#endif14
   return s;
  }
 
 
}
