/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd.policy;

import java.lang.reflect.Method;


/**
 * The common superclass of all dispatching policies.
 * <p>
 * The methods defined in this class can be
 * overridden in order to create user-defined dispatching policies.
 * Please check out the invidual methods below
 * to know more.
 * 
 * @version $Revision: 1.10 $, $Date: 2005/04/03 00:44:23 $
 * @author Antonio Cunei
 *
 */
abstract public class Dispatching {

 /**
  * Returns an instance of this dispatching policy.<p>
  * This method must be defined in every user-defined
  * dispatching policy. It must return one instance of the
  * same class, normally a singleton allocated during construction.
  * This method is used reflexively to allow access to the
  * fields and methods of the class given the class name.
  * 
  * @return an instance of this class (a singleton)
  */
  public static Dispatching theDispatcher() { return null; }
 

  /**
   * Remaps the arguments when an unexpected null argument is
   * encountered.
   * If an argument is null, and it is not caught by
   * <code>@IfNull</code>, find a suitable alternative
   * combination of classes here.<p>
   * Define this method if you would like to be able to
   * handle null arguments in a more general form with
   * respect to what is allowed by <code>@IfNull</code> tags.
   * While <code>@IfNull</code> operate on individual
   * arguments, this method can be used to alter the interpretation
   * of all the arguments together when one of them is
   * unexpectedly null.<p>
   * As an alternative to this general method, you can use
   * the simpler form {@link #remapNull(Class[], Method[])}.
   * If both are overridden, the most general form will be used.
   * For maximum efficiency, do not override <code>remapNull</code>,
   * but use the simpler <code>@IfNull</code> tag instead.
   * @see ovm.polyd.tags.IfNull
   * @param args the array of classes of non-raw arguments
   * @param meth the array of methods that will be used for the selection
   * @param raws the array of raw arguments
   * @param implems the array of implementing bodies
   * @param proto the prototype through which this call was made
   * @param name if the prototype was renamed, this is the new name
   *  (otherwise, the same as the prototype)
   * @return an array containing exactly two elements. The first
   *  is the remapped array of classes, the second is the
   *  remapped array of raw arguments
   */
 //
 // DO NOT CHANGE THIS METHOD IN THIS CLASS. ClassBuilder optimizes if
 // it sees that these method is not overridden, and does not add an
 // actual call to remapNull(), leaving instead the standard
 // dispatching routine to merrily fail with the same exception.
 //
 public Object[][] remapNull(Class[] args, Method[] meth,Object[] raws,
                        Class[] implems,Method proto,String name) {
   return new Object[][]{remapNull(args,meth),raws};
 }

 // The second (simpler) form just remaps the classes of non-raw args
 /**
  * Remaps the arguments when an unexpected null argument is
  * encountered.<p>
  * If an argument is null, and it is not caught by
  * <code>@IfNull</code>, find a suitable alternative
  * combination of classes here.
  * @see #remapNull(Class[], Method[], Object[], Class[], Method, String)
  * @param args the array of classes of non-raw arguments
  * @param meth the array of methods that will be used for the selection
  * @return the remapped array of classes
  */
 // DO NOT CHANGE THIS METHOD IN THIS CLASS. ClassBuilder optimizes if
 // it sees that these method is not overridden, and does not add an
 // actual call to remapNull(), leaving instead the standard
 // dispatching routine to merrily fail with the same exception.
 //
 public Class[] remapNull(Class[] args, Method[] meth) {
   throw new NullPointerException();
 }

 //---------------------

 /**
  * Dynamically selects the most appropriate method for a
  * certain combination of arguments.<p>
  * The most crucial part of a dispatching policy, this
  * method contains the logic required to decide which is
  * the implementing method that better fits a certain
  * calling context. In order to help in the choice,
  * PolyD supplies various bits of information.<p>
  * It is assumed that the same combination of arguments
  * always leads to the same method being selected. PolyD
  * relies on this property in order to cache the method
  * selection, and you should not assume that <code>bestMatch</code>
  * will be called every time a message is dispatched.<p>
  * If your particular dispatching policy cannot guarantee
  * that it will select always the same method for a given
  * combination of arguments, or if you absolutely need to
  * make sure that <code>bestMatch</code> will be called
  * every single time a message is dispatched, then you
  * will have to disable caching altogether for this
  * policy, see {@link #disableCaching()}.<p>
  * If you do not require raw arguments, or the whole
  * set of arguments available to this method, a
  * simpler form can be defined instead, see
  * {@link #bestMatch(Class[], Method[])}.
  * 
  * @param args the array of classes of non-raw arguments
  * @param meth the array of methods involved in the selection
  * @param raws the array of raw arguments
  * @param implems the array of implementing bodies
  * @param proto the prototype through which the message was sent
  * @param name if the prototype was renamed, this is the new name
  * @return the index of the best applicable method in the
  *  <code>meth</code> array, or -1 if none can be chosen.
  */
 // return the index, within meth, which represents the best
 // combination for this combination of args, that is, the most
 // specific
 public int bestMatch(Class[] args,Method[] meth,Object[] raws,
                        Class[] implems,Method proto,String name) {

   return bestMatch(args,meth);
 }
 /**
  * Dynamically selects the most appropriate method for a
  * certain combination of arguments.<p>
  * @see #bestMatch(Class[], Method[], Object[], Class[], Method, String)
  * @param args the array of classes of non-raw arguments
  * @param meth the array of methods involved in the selection
  * @return the index of the best applicable method in the
  *  <code>meth</code> array, or -1 if none can be chosen.
  */
 public int bestMatch(Class[] args, Method[] meth) {
  return -1;
 }


 /**
  * Performs a static preselection on a set of methods, and/or
  * a consistency check.<p>
  * Redefine this method if you would like to perform a static
  * analysis on the set of methods that might be called through
  * each prototype defined in the dispatcher interface.
  * The main uses of this method are:<ol>
  * <li>Verifying that the set of methods is internally consistent,
  * for example checking that there are no duplicates or ambiguities.
  * <li>Performing a static preselection, determining the subset
  * of methods that can really be ever chosen as a result of a
  * message sent through the given prototype.
  * </ol>
  * If an inconsistency is detected, you should throw some sort
  * of <code>PolicyException</code>. Return the applicable subset
  * of <code>meth</code> (or all of it). Redefining this method
  * in a user-defined policy is not mandatory, but it is
  * recommended. It is also possible to define in its place
  * the simpler {@link #compatibleSet(Class[], Method[])}.
  * 
  * @see ovm.polyd.PolicyException
  * @param args the array of classes of non-raw arguments
  * @param meth the array of methods that will be used for the selection
  * @param implems the array of implementing bodies
  * @param proto the prototype through which the methods might be called
  * @param name if the prototype was renamed, this is the new name
  * @return the selected subset, or <code>meth</code>
  */
 public Method[] compatibleSet(Class[] args,Method[] meth,Class[] implems,Method proto,String name) {
  return compatibleSet(args,meth);
 }

/**
 * Performs a static preselection on a set of methods, and/or
 * a consistency check.
 * 
 * @see #compatibleSet(Class[], Method[], Class[], Method, String)
 * @param args the array of classes of non-raw arguments
 * @param meth the array of methods that will be used for the selection
 * @return the selected subset, or <code>meth</code>
 */
 public Method[] compatibleSet(Class[] args, Method[] meth) {
  return meth;
 }
 
 /**
  * Handles messages which cannot be dispatched to any method,
  * according to the current dispatching policy.<p>
  * This method can take whatever action might be appropriate,
  * including calling further methods. However, it should not
  * throw a <code>MissingMethodException</code>. If no
  * reasonable corrective action can be taken, simply
  * return <code>false</code> to the caller.<p>
  * An alternative, simpler form of this method is also
  * available, see {@link #handleMissing(Class[], Method[])}.
  * Redefining <code>remapNull</code> in user-defined policies
  * is not required.
  * 
  * @param args the array of classes of non-raw arguments
  * @param meth the array of methods involved in the selection
  * @param raws the array of raw arguments
  * @param implems the array of implementing bodies
  * @param proto the prototype through which the methods might be called
  * @param name if the prototype was renamed, this is the new name
  * @return <code>true</code> if it was possible to handle the message, <code>false</code> otherwise
  */
 public boolean handleMissing(Class[] args,Method[] meth,Object[] raws,
                        Class[] implems,Method proto,String name) {

  return handleMissing(args,meth);
 }

 /**
  * Handles messages which cannot be dispatched to any method,
  * according to the current dispatching policy.
  * @see #handleMissing(Class[], Method[], Object[], Class[], Method, String)
  * @param args the array of classes of non-raw arguments
  * @param meth the array of methods involved in the selection
  * @return <code>true</code> if it was possible to handle the message, <code>false</code> otherwise
  */
 public boolean handleMissing(Class[] args, Method[] meth) {
  return false;
 }

/**
 * Disables method caching for this policy.<p>
 * If this method is overridden in your user-defined policy,
 * and it returns <code>true</code>, then the whole caching mechanism
 * offered by PolyD will be disabled for this dispatching
 * policy. As a result, <code>bestMatch</code> will be
 * called each and every time a message is dispatched.<p>
 * Disabling caching involves a large performance penalty,
 * and should really only be used for those policies
 * for which the method selection depends on external
 * factors and the chosen method may be different from
 * time to time even for the same arguments. Alternatively,
 * disabling caching can sometimes be of use during debugging.<p>
 * User-defined policies shuld not normally override this
 * method. Note that this method should always return
 * the same value for one given dispatching policy
 * (either always <code>true</code>, or always <code>false</code>).
 * @return <code>true</code> if method caching
 *  must be disabled, <code>false</code> otherwise
 */
 public boolean disableCaching() { return false; }
 
 /**
  * Returns the name of this dispatcher
  * @return a string containing the name of the dispatcher
  */
 public String toString() { return "<Unknown>"; }
 
 // constructor
 protected Dispatching() {}
}
