/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd.policy;
import java.lang.reflect.Method;

import ovm.polyd.PolicyException;
import ovm.polyd.Utils;
import ovm.polyd.Utils.Diff;
import static ovm.polyd.Utils.Diff.*;

/**
 * This class implements multidispatching, as a PolyD dispatching policy.
 * All possible ambiguities are detected, most of them at
 * dispatcher-building time, and the rest dynamically.
 * 
 * @version $Revision: 1.11 $, $Date: 2005/04/02 07:42:32 $
 * @author Antonio Cunei
 *
 */
public class MultiDisp extends Dispatching {
 protected MultiDisp() {}
 private static final Dispatching theDispatcher=new MultiDisp();
 /**
  * Returns an instance of this dispatching policy.
  * 
  * @return an instance of this policy (a singleton)
  */
 public static Dispatching theDispatcher() { return theDispatcher; }

 /**
  * Thrown by MultiDisp if an error condition is detected.
  */
 public static class MultiDispException extends PolicyException {
  MultiDispException(String s) {super(s);}
 }

	 public Method[] compatibleSet(Class[] args,Method[] meth) {
	  
	  if (meth.length==0) return meth; // should never happen
	  
	  int ar=meth[0].getParameterTypes().length; // arity
	  if (ar==0) {
	   if (meth.length!=1) {
	    throw new MultiDispException("Duplicated method: "+meth[0].getName()+"()\n");
	   } else return meth;
	  }
	  
	  boolean fault=false; // let's scan all, and stop at the end
   String troubleString="";
	  int i,j,q;

   // rather inefficient algorithm, but efficiency is not
   // terribly important within this class
	  Class[] toSearchP=new Class[ar];

	  for (i=0;i<meth.length;i++) {
	   Method oneM=meth[i];
	   Class[] one=oneM.getParameterTypes();
	   for (j=i+1;j<meth.length;j++) {
	    Method twoM=meth[j];
	    Class[] two=twoM.getParameterTypes();
	    
	    Diff direction=Utils.classArrayDiff(one,two,toSearchP);
	    
	    if (direction==Equal) {
      troubleString+="Duplicated method: "+oneM.getName()+"("+toSearchP[0].getName();
	     for (q=1;q<ar;troubleString+=","+toSearchP[q++].getName());
      troubleString+=").";
      troubleString+="  It appears in "+oneM.getDeclaringClass().getName()+" and";
      troubleString+="             in "+twoM.getDeclaringClass().getName()+"\n";
	     fault=true;
	    }
	    
	    if (direction==Crossed) {  // there might be an ambiguity: we shall check that
	     // the combination in toSearchP is present somewhere
	     // (cubic in the worst case. This code should be made more efficient!)
	     if (Utils.isThere(toSearchP,meth)<0) {
       troubleString+="An ambiguity was detected.";
	      System.out.print("The methods: "+oneM.getName()+"("+one[0].getName());
	      for (q=1;q<ar;troubleString+=","+one[q++].getName());
       troubleString+=")";
       troubleString+="    defined in "+oneM.getDeclaringClass().getName()+",";
       troubleString+="        and: "+twoM.getName()+"("+two[0].getName();
	      for (q=1;q<ar;troubleString+=","+two[q++].getName());
       troubleString+=")\n    defined in "+twoM.getDeclaringClass().getName();
       troubleString+="        are conflicting. To solve the";
       troubleString+=" ambiguity, define:";
       troubleString+="             "+twoM.getName()+"("+toSearchP[0].getName();
	      for (q=1;q<ar;troubleString+=","+toSearchP[q++].getName());
       troubleString+=").\n";
	      // I could potentially also suggest a suitable return type
	      // looking at the list of all the prototypes compatible with the
	      // newly discovered and missing multimethod.
	      fault=true;
	     }
	    }
	    
	   }
	  }
   if (fault)
    throw new MultiDispException(troubleString);
   else
    return meth;
	 }
	 
	 
	 
	 // return the index, within meth, which represents the best
	 // combination for this combination of args, that is, the most
	 // specific
	 public int bestMatch(Class[] args,Method[] meth) {


   Method[] mm=new Method[meth.length];

   int scan,count=0;
   for (scan=0;scan<meth.length;scan++) {
    if (Utils.isEqualOrSubArrayNull(args,meth[scan].getParameterTypes())) {
     mm[count++]=meth[scan];
    }
   }

   // new ambiguities might have been introduced by the dynamic loading!
   // I must check again (uses a quadratic algorithm)
   
   // elements from 0 to scan-1
   int q,z;
   Diff res;
   outer:
   for (q=count-1;q>0;q--) {
    if (mm[q]==null) continue;
    for (z=q-1;z>=0;z--) {
     if (mm[z]==null) continue;
     // let's try to remove one of the two
     res=Utils.classArrayDiff(mm[q].getParameterTypes(),mm[z].getParameterTypes());
     if (res==Equal) {
      throw new MultiDispException(
       "Policy internal error: bestMatch in MultiDisp\n"+
       "found duplicate method class arrays."
      );
     } else if (res==Sub) {
      mm[z]=null;
     } else if (res==Super) {
      mm[q]=null;
      continue outer;
     }
    }
   }

   // There should only be one survivor. if >1 -> ambiguitiy, if zero internal error.

   int index=0;
   int total=0;
   
   for (scan=0;scan<count;scan++) {
    if (mm[scan]!=null) {
     index=scan;
     total++;
    }
   }

   if (total==0) {
      throw new MultiDispException(
       "Policy internal error: bestMatch in MultiDisp\n"+
       "did not find an applicable method."
      );
   }
   
   if (total>1) {
    String s="Ambiguity detected while using multiple dispatching.\n";
    s+="The following combination of classes:\n  (";
    String sep="";
    for (Class c : args) {
     s+=sep+c.getName();
     sep=",";
    }
    s+=")\n";
    s+="is ambiguous for the following methods:";
    for (scan=0;scan<count;scan++) {
     if (mm[scan]!=null) {
      s=s+"\n"+mm[scan].getName()+"(";
      sep="";
      for (Class c : mm[scan].getParameterTypes()) {
       s+=sep+c.getName();
       sep=",";
      }
      s+="), defined in "+mm[scan].getDeclaringClass().getName();
     }
    }
    throw new MultiDispException(s);
   }
   
    for (scan=0;scan<meth.length;scan++) {
     if (mm[index]==meth[scan]) return scan;
    }

   return -1; // impossible.
   
  }
  
  public String toString() { return "MultiDisp"; }
}
