/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd.policy;
import java.lang.reflect.Method;

import ovm.polyd.PolicyException;
import ovm.polyd.Utils;
import ovm.polyd.Utils.Diff;
import ovm.polyd.policy.MultiDisp.MultiDispException;
import static ovm.polyd.Utils.Diff.*;

/**
 * This class implements a non-subsumptive dispatcher, as a PolyD dispatching policy.
 * Only methods whose list of parameters matches exactly the runtime class of the
 * message arguments will be called.
 * 
 * @version $Revision: 1.7 $, $Date: 2005/04/02 07:42:32 $
 * @author Antonio Cunei
 *
 */
public class NonSubsump extends Dispatching {
 protected NonSubsump() {}
 private static final Dispatching theDispatcher=new NonSubsump();
 /**
  * Returns an instance of this dispatching policy.
  * 
  * @return an instance of this policy (a singleton)
  */
 public static Dispatching theDispatcher() { return theDispatcher; }

 /**
  * Thrown by NonSubsump if an error condition is detected.
  */
 public static class NonSubsumpException extends PolicyException {
  NonSubsumpException(String s) {super(s);}
 }

 public Method[] compatibleSet(Class[] args,Method[] meth) {

	  if (meth.length==0) return meth; // should never happen
	  
	  int ar=meth[0].getParameterTypes().length; // arity
   if (ar==0) {
    if (meth.length!=1) {
     throw new NonSubsumpException("Duplicated method: "+meth[0].getName()+"()\n");
    } else return meth;
   }
	  
	  // we only need to check for duplicates
	  
	  boolean fault=false; // let's scan all, and stop at the end
   String troubleString="";
	  int i,j,q;

	  Class[] toSearchP=new Class[ar];

	  for (i=0;i<meth.length;i++) {
	   Method oneM=meth[i];
	   Class[] one=oneM.getParameterTypes();
	   for (j=i+1;j<meth.length;j++) {
	    Method twoM=meth[j];
	    Class[] two=twoM.getParameterTypes();
	    
	    Diff direction=Utils.classArrayDiff(one,two,toSearchP);
	    
     if (direction==Equal) {
      troubleString+="Duplicated method: "+oneM.getName()+"("+toSearchP[0].getName();
      for (q=1;q<ar;troubleString+=","+toSearchP[q++].getName());
      troubleString+=").";
      troubleString+="  It appears in "+oneM.getDeclaringClass().getName()+" and";
      troubleString+="             in "+twoM.getDeclaringClass().getName()+"\n";
      fault=true;
     }
	   }
	  }
   if (fault)
    throw new NonSubsumpException(troubleString);
   else
    return meth;
	 }
	 

 
 
 public int bestMatch(Class[] args,Method[] meth) {
	  
	  if (meth.length==0) return -1;
	  
	  int ar=args.length; // arity
	  boolean maybe;
	  int scan,q;
	  Class[] params;
	  Class[] best=new Class[ar];
	  int bestIndex=-1;
	  
	  if (ar==0) return 0; // only one body
	  
	  for (scan=0;scan<meth.length;scan++) {
	   
	   params=meth[scan].getParameterTypes();
	   for (q=0,maybe=true;q<ar;q++) {
	    if (Utils.classDiff(args[q],params[q])!=Equal) {maybe=false; break;}
	   }
	   
	   if (maybe) return scan;
	  }
	  
	  // returns -1 if no match.
	  
	  return -1;
	 }
	 

  public String toString() { return "NonSubsump"; }
  
}
