/*
  (c) 2005 Antonio Cunei, All rights reserved.

  This file is part of PolyD.

  PolyD is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  PolyD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PolyD; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package ovm.polyd.policy;
import java.lang.reflect.Method;

import ovm.polyd.InternalError;
import ovm.polyd.DispatcherCreationException;
import ovm.polyd.PolyDException;
import ovm.polyd.Utils;
import ovm.polyd.Utils.Diff;
import ovm.polyd.policy.MultiDisp.MultiDispException;
import static ovm.polyd.Utils.Diff;
import static ovm.polyd.Utils.Diff.*;

/**
 * This class implements overloading, as a PolyD dispatching policy.
 * All  ambiguities are detected statically.
 * 
 * @version $Revision: 1.8 $, $Date: 2005/04/02 07:42:32 $
 * @author Antonio Cunei
 *
 */
public class Overloading extends Dispatching {
 protected Overloading() {}
 private static final Dispatching theDispatcher=new Overloading();
 /**
  * Returns an instance of this dispatching policy.
  * 
  * @return an instance of this policy (a singleton)
  */
 public static Dispatching theDispatcher() { return theDispatcher; }

 /**
  * Thrown by Overloading if an error condition is detected.
  */
 public static class OverloadingException extends PolyDException {
  OverloadingException(String s) {super(s);}
 }
	
 public int bestMatch(Class[] args,Method[] meth) {
	  if (meth.length!=1) {
      throw new OverloadingException(
       "Policy internal error: Overloading policy was asked to decide\n"+
       "  dynamically among multiple methods, or none."
      );
   }
   return 0;
	}
  

 public Method[] compatibleSet(Class[] args, Method[] meth) {

   if (meth.length==0) return meth; // should never happen
   
   int ar=meth[0].getParameterTypes().length; // arity
   if (ar==0) {
    if (meth.length!=1) {
     throw new OverloadingException("Duplicated method: "+meth[0].getName()+"()\n");
    } else return meth;
   }
   
   Method[] mm=new Method[meth.length];

   int scan,count;
   for (scan=0,count=0;scan<meth.length;scan++,count++) {
    if (Utils.isEqualOrSubArrayNull(args,meth[scan].getParameterTypes())) {
     mm[count]=meth[scan];
    }
   }

   // discover ambiguities (uses a quadratic algorithm)
   
   // elements from 0 to scan-1
   int q,z;
   Diff res;
   outer:
   for (q=count-1;q>0;q--) {
    if (mm[q]==null) continue;
    for (z=q-1;z>=0;z--) {
     if (mm[z]==null) continue;
     // let's try to remove one of the two
     res=Utils.classArrayDiff(mm[q].getParameterTypes(),mm[z].getParameterTypes());
     if (res==Equal) {
      String troubleString="";
      troubleString+="Duplicated method: "+mm[q].getName()+"("+mm[q].getParameterTypes()[0].getName();
      for (int qq=1;qq<ar;troubleString+=","+mm[q].getParameterTypes()[qq++].getName());
      troubleString+=").";
      troubleString+="  It appears in "+mm[q].getDeclaringClass().getName()+" and";
      troubleString+="             in "+mm[z].getDeclaringClass().getName()+"\n";
      throw new OverloadingException(troubleString);
     } else if (res==Sub) {
      mm[z]=null;
     } else if (res==Super) {
      mm[q]=null;
      continue outer;
     }
    }
   }

   // There should only be one survivor. if >1 -> ambiguity, if zero error (it should be impossible)

   int index=0;
   int total=0;
   
   for (scan=0;scan<count;scan++) {
    if (mm[scan]!=null) {
     index=scan;
     total++;
    }
   }

   if (total==0) {
      throw new OverloadingException(
       "compatibleSet() in the Overloading policy\n"+
       "did not find an applicable method."
      );
   }
   
   if (total>1) {
    String s="Ambiguity detected while using overloading.";
    s+="The following combination of classes:\n  (";
    String sep="";
    for (Class c : args) {
     s+=sep+c.getName();
     sep=",";
    }
    s+=")\n";
    s+="is ambiguous for the following methods:";
    for (scan=0;scan<count;scan++) {
     if (mm[scan]!=null) {
      s=s+"\n"+mm[scan].getName()+"(";
      sep="";
      for (Class c : mm[scan].getParameterTypes()) {
       s+=sep+c.getName();
       sep=",";
      }
      s+="), defined in "+mm[scan].getDeclaringClass().getName();
     }
    }
    throw new OverloadingException(s);
   }
   
    for (scan=0;scan<meth.length;scan++) {
     if (mm[index]==meth[scan]) {
      Method[] m=new Method[1];
      m[0]=meth[scan];
      return m;
     }
    }

   return new Method[0]; // impossible.
  }
  
  public String toString() { return "Overloading"; }
}
